
from pycamia import info_manager

__info__ = info_manager(
    project = "PyZMyc",
    package = "pyoverload",
    fileinfo = "Useful tools for decorators."
)

__all__ = """
    raw_function
    return_type_wrapper
    decorator
""".split()

import sys
from functools import wraps

def raw_function(func):
    if hasattr(func, "__func__"):
        return func.__func__
    return func

def _get_wrapped(f):
    while hasattr(f, '__wrapped__'): f = f.__wrapped__
    return f

def decorator(wrapper_func):
    if not callable(wrapper_func): raise TypeError("@decorator wrapping a non-wrapper")
    def wrapper(*args, **kwargs):
        if not kwargs and len(args) == 1:
            func = args[0]
            raw_func = raw_function(func)
            if callable(raw_func):
                func_name = f"{raw_func.__name__}[{wrapper_func.__qualname__.split('.')[0]}]"
                wrapped_func = wraps(raw_func)(wrapper_func(raw_func))
                wrapped_func.__name__ = func_name
                wrapped_func.__doc__ = raw_func.__doc__
                # return wrapped_func
                if 'staticmethod' in str(type(func)): trans = staticmethod
                elif 'classmethod' in str(type(func)): trans = classmethod
                else: trans = lambda x: x
                return trans(wrapped_func)
        return decorator(wrapper_func(*args, **kwargs))
    return wraps(wrapper_func)(wrapper)

def _mid(x): return x[1] if len(x) > 1 else x[0]
def _rawname(s): return _mid(str(s).split("'"))

stack_error = lambda x: TypeError(f"Unexpected function stack for {x}, please contact the developer for further information. ")

def _get_frames():
    frames = []
    frame = sys._getframe()
    fname = frame.f_back.f_code.co_name
    while frame is not None:
        frame_file = _rawname(frame)
        if frame_file.startswith('<') and frame_file.endswith('>') and frame_file != '<stdin>':
            frame = frame.f_back
            continue
        frames.append(frame)
        if len(frames) >= 4: return frames[2:]
        frame = frame.f_back
    raise stack_error(fname)

def get_environ_locals():
    _, client_frame = _get_frames()
    return client_frame.f_locals

def get_environ_globals():
    _, client_frame = _get_frames()
    return client_frame.f_globals




from .manager import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = '<main>',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.0.24',
    contact = 'bertiezhou@163.com',
    keywords = ['environment', 'path', 'touch'],
    description = 'The main package and a background support of project PyCAMIA. ',
    requires = []
).check()

from .environment import get_environ_locals, get_environ_globals, get_args_expression #*
from .exception import touch, crashed, avouch, Error #*
from .functions import empty_function, const_function, identity_function #*
from .inout import no_print, str_print, SPrint #*
from .manager import info_manager #*
from .timing import time_this, Timer, Jump, scope, jump, Workflow, periodic #*
from .listop import prod, argmin, argmax, flat_list, item, to_list, to_tuple #*
from .strop import str_len, str_slice, find_all, sorted_dict_repr, enclosed_object, tokenize #*

















































































































from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "batorch",
    fileinfo = "The inherited tensor from 'torch' with batch.",
    requires = "torch"
)

__all__ = """
    Device
    DeviceCPU
    Tensor
    Size
    set_autodevice
    unset_autodevice
    is_autodevice
""".split()

if __name__ == '__main__':
    pass




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File containing commonly used similarity measures in medical image analysis. ",
    help = "Use `metric['ABBR.'](A, B)` to compute the similarity.",
    requires = "SimpleITK"
).check()

__all__ = """
    MutualInformation
    NormalizedMutualInformation
    KLDivergence
    CorrelationOfLocalEstimation
    NormalizedVectorInformation
    Cos2Theta
    SumSquaredDifference
    MeanSquaredErrors
    PeakSignalToNoiseRatio
    CrossEntropy
    CrossCorrelation
    NormalizedCrossCorrelation
    StructuralSimilarity
    Dice
    LabelDice
""".split()

with __info__:
    import torch
    import batorch as bt
    import SimpleITK as sitk

######### Section 1: Information Based ########
eps = 1e-6

def Bspline(i, U):
    i = bt.tensor(i); U = bt.tensor(U)
    return (
        bt.where(i == -1, (1 - U) ** 3 / 6,
        bt.where(i == 0, U ** 3 / 2 - U * U + 2 / 3,
        bt.where(i == 1, (- 3 * U ** 3 + 3 * U * U + 3 * U + 1) / 6,
        bt.where(i == 2, U ** 3 / 6,
        bt.zeros_like(U)))))
    )

def dBspline(i, U):
    i = bt.tensor(i); U = bt.tensor(U)
    return (
        bt.where(i == -1, - 3 * (1 - U) ** 2 / 6,
        bt.where(i == 0, 3 * U ** 2 / 2 - 2 * U,
        bt.where(i == 1, (- 3 * U ** 2 + 2 * U + 1) / 2,
        bt.where(i == 2, 3 * U ** 2 / 6,
        bt.zeros_like(U)))))
    )

def dBspline_WRT_I1(i, U):
    '''
    THe derivative of Bspline function with respect to I2.
    i, U: n_batch x n_hist x n_data
    '''
    return dBspline(i[:, 0], U[:, 0]) * Bspline(i[:, 1], U[:, 1])

def dBspline_WRT_I2(i, U):
    '''
    THe derivative of Bspline function with respect to I2.
    i, U: n_batch x n_hist x n_data
    '''
    return Bspline(i[:, 0], U[:, 0]) * dBspline(i[:, 1], U[:, 1])

class JointHistogram(bt.autograd.Function):

    @staticmethod
    def forward(ctx, I1, I2, nbin=100):
        with bt.no_grad():
            if hasattr(ctx, 'JH'): del ctx.JH
            nbin = bt.tensor(nbin)
            data_pair = bt.stack(I1.flatten(1), I2.flatten(1), dim={1})
            nbatch, nhist, ndata = data_pair.ishape
            indices = []; values = []
            ctx.window = (bt.image_grid(4, 4) - 1).flatten(1).transpose(0, 1)
            for shift in ctx.window:
                # [nbatch] x {nhist} x ndata
                hist_pos = data_pair * nbin
                index = bt.clamp(bt.floor(hist_pos).long() + shift, 0, nbin - 1)
                batch_idx = bt.arange(nbatch).expand_to([nbatch], {1}, ndata)
                index = bt.cat(batch_idx, index, 1)
                value = Bspline(shift.expand_to(data_pair), bt.decimal(hist_pos)).prod(1)
                indices.append(index)
                values.append(value)
            # n_batch x (1 + n_hist) x (n_data x 4 ** n_hist)
            Mindices = bt.cat(indices, -1)
            # n_batch x (n_data x 4 ** n_hist)
            Mvalues = bt.cat(values, -1)
            # (1 + n_hist) x (n_batch x n_data x 4 ** n_hist)
            indices = Mindices.transpose(0, 1).flatten(1)
            # (n_batch x n_data x 4 ** n_hist)
            values = Mvalues.flatten(0)
            if bt.Device == bt.DeviceCPU: creator = torch.sparse.FloatTensor
            else: creator = torch.cuda.sparse.FloatTensor
            collected = creator(indices, values, (nbatch, nbin, nbin)).to_dense()
            collected = bt.Tensor(collected, batch_dim=0)

            ctx.nbin = nbin
            ctx.Ishape = I1.shape
            ctx.data_pair = data_pair
            ctx.JH = collected / ndata
        return ctx.JH

    @staticmethod
    def backward(ctx, grad_output):
        with bt.no_grad():
            nbin = ctx.nbin
            data_pair = ctx.data_pair
            nbatch, nhist, ndata = data_pair.ishape
            dPdI1 = bt.zeros(ctx.Ishape)
            dPdI2 = bt.zeros(ctx.Ishape)
            for shift in ctx.window:
                # [nbatch] x {nhist} x ndata
                shift = shift.view(1, 2, 1)
                hist_pos = data_pair * nbin
                index = torch.clamp(torch.floor(hist_pos).long() + shift, 0, nbin - 1)
                grad_y = grad_output[(slice(None),) + index.split(1, 1)].squeeze(2)
                value = grad_y.gather(0, bt.arange(nbatch).long().unsqueeze(0).unsqueeze(-1).repeat(1, 1, ndata)).view(ctx.Ishape)
                dPdI1 += value * dBspline_WRT_I1(shift, bt.decimal(data_pair * nbin)).view(ctx.Ishape)
                dPdI2 += value * dBspline_WRT_I2(shift, bt.decimal(data_pair * nbin)).view(ctx.Ishape)
        return dPdI1, dPdI2, None

def MutualInformation(A, B, nbin=100):
    assert A.has_batch and B.has_batch
    Pab = JointHistogram.apply(A, B, nbin)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    Hxy = - bt.sum(Pab * bt.log2(bt.where(Pab < eps, bt.ones_like(Pab), Pab)), [1, 2])
    Hx = - bt.sum(Pa * bt.log2(bt.where(Pa < eps, bt.ones_like(Pa), Pa)), 1)
    Hy = - bt.sum(Pb * bt.log2(bt.where(Pb < eps, bt.ones_like(Pb), Pb)), 1)
    return Hx + Hy - Hxy

def NormalizedMutualInformation(A, B, nbin=100):
    assert A.has_batch and B.has_batch
    Pab = JointHistogram.apply(A, B, nbin)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    Hxy = - bt.sum(Pab * bt.log2(bt.where(Pab < eps, bt.ones_like(Pab), Pab)), [1, 2])
    Hx = - bt.sum(Pa * bt.log2(bt.where(Pa < eps, bt.ones_like(Pa), Pa)), 1)
    Hy = - bt.sum(Pb * bt.log2(bt.where(Pb < eps, bt.ones_like(Pb), Pb)), 1)
    return (Hx + Hy) / Hxy

def KLDivergence(A, B, nbin=100):
    assert A.has_batch and B.has_batch
    Pab = JointHistogram.apply(A, B, nbin)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    return (Pa * bt.log2(bt.where(Pb < eps, bt.ones_like(Pa), Pa / Pb.clamp(min=eps)).clamp(min=eps))).sum(1)

###############################################

######## Section 2: Cross Correlation #########

def local_matrix(A, B, s=0, kernel="Gaussian", kernel_size=3):
    if isinstance(kernel, str):
        if kernel.lower() == "gaussian": kernel = bt.gaussian_kernel(n_dims = A.nspace, kernel_size = kernel_size).unsqueeze(0, 0)
        elif kernel.lower() == "mean": kernel = bt.ones(*(kernel_size,) * A.nspace).unsqueeze(0, 0) / (kernel_size ** A.nspace)
    elif hasattr(kernel, 'shape'): kernel_size = kernel.size(-1)

    def mean(a):
        op = eval("bt.nn.functional.conv%dd"%A.nspace)
        if a.has_batch: x = a.unsqueeze({1})
        else: x = a.unsqueeze([0], {1})
        return op(x, kernel, padding = kernel_size // 2).squeeze(*((1,) if a.has_batch else (0, 0)))

    if s > 0:
        GA = bt.grad_image(A)
        GB = bt.grad_image(B)
        point_estim = bt.stack(bt.dot(GA, GA), bt.dot(GA, GB), bt.dot(GB, GB), dim={int(A.has_batch)})
    else: point_estim = 0

    MA = mean(A)
    MB = mean(B)
    local_estim = bt.stack(mean(A * A) - MA ** 2, mean(A * B) - MA * MB, mean(B * B) - MB ** 2, dim={int(A.has_batch)})

    return s * point_estim + local_estim

def CorrelationOfLocalEstimation(A, B, s=0, kernel="Gaussian", kernel_size=3):
    assert A.has_batch and B.has_batch
    S11, S12, S22 = local_matrix(A, B, s=s, kernel=kernel, kernel_size=kernel_size).split()
    return (bt.divide(S12 ** 2, S11 * S22, tol=eps).squeeze(1) + eps).sqrt().mean()

###############################################

########## Section 3: Local Gradient ##########

def NormalizedVectorInformation(A, B):
    assert A.has_batch and B.has_batch
    GA = bt.grad_image(A)
    GB = bt.grad_image(B)
    return bt.divide(bt.dot(GA, GB) ** 2, bt.dot(GA, GA) * bt.dot(GB, GB), tol=eps).mean()

def Cos2Theta(A, B):
    assert A.has_batch and B.has_batch
    GA = bt.grad_image(A)
    GB = bt.grad_image(B)
    return bt.divide(bt.dot(GA, GB) ** 2, bt.dot(GA, GA) * bt.dot(GB, GB), tol=eps)

###############################################

####### Section 4: Intensity Difference #######

def SumSquaredDifference(A, B):
    assert A.has_batch and B.has_batch
    return ((A - B) ** 2).sum()

def MeanSquaredErrors(A, B):
    assert A.has_batch and B.has_batch
    return ((A - B) ** 2).mean()

def PeakSignalToNoiseRatio(A, B):
    assert A.has_batch and B.has_batch
    return 10 * bt.log10(bt.max((A.max(), B.max())) ** 2 / ((A - B) ** 2).mean())

###############################################

##### Section 5: Distribution Similarity ######

def CrossEntropy(y, label):
    assert y.has_batch and label.has_batch and y.has_channel and label.has_channel
    ce = - label * bt.log(y.clamp(1e-10, 1.0))
    return ce.sum(ce.channel_dimension).mean()

def CrossCorrelation(A, B):
    assert A.has_batch and B.has_batch
    dA = A - A.mean(); dB = B - B.mean()
    return (dA * dB).sum()

def NormalizedCrossCorrelation(A, B):
    assert A.has_batch and B.has_batch
    dA = A - A.mean(); dB = B - B.mean()
    return (dA * dB).sum() / (dA ** 2).sum().sqrt() / (dB ** 2).sum().sqrt()

def StructuralSimilarity(A, B, k1=0.01, k2=0.03):
    assert A.has_batch and B.has_batch
    varA = ((A - A.mean()) ** 2).mean()
    varB = ((B - B.mean()) ** 2).mean()
    covAB = ((A - A.mean()) * (B - B.mean())).mean()
    L = bt.max((A.max(), B.max()))
    c1, c2 = k1 * L, k2 * L
    num = (2 * A.mean() * B.mean() + c1 ** 2) * (2 * covAB + c2 ** 2)
    den = (A.mean() ** 2 + B.mean() ** 2 + c1 ** 2) * (varA + varB + c2 ** 2)
    return num / den

###############################################

########## Section 6: Region Overlap ##########

def Dice(A, B, multi_label = False):
    '''
    if multi_label:
        A: (n_batch, n_label, n_1, n_2, ..., n_k)
        B: (n_batch, n_label, n_1, n_2, ..., n_k)
        return: (n_batch, n_label)
    else:
        A: (n_batch, n_1, n_2, ..., n_k)
        B: (n_batch, n_1, n_2, ..., n_k)
        return: (n_batch,)
    '''
    assert A.has_batch and B.has_batch
    ABsum = A.sum() + B.sum()
    return 2 * (A * B).sum() / (ABsum + eps)

def LabelDice(A, B, class_labels=None):
    '''
    :param A: (n_batch, n_1, ..., n_k)
    :param B: (n_batch, n_1, ..., n_k)
    :param class_labels: list[n_class]
    :return: (n_batch, n_class)
    '''
    assert A.has_batch and B.has_batch
    if not class_labels: class_labels = sorted(A.unique().tolist() + B.unique().tolist())
    A_labels = [1 - bt.clamp(bt.abs(A - i), 0, 1) for i in class_labels]
    B_labels = [1 - bt.clamp(bt.abs(B - i), 0, 1) for i in class_labels]
    A_maps = bt.stack(A_labels, {1})
    B_maps = bt.stack(B_labels, {1})
    return Dice(A_maps, B_maps)

###############################################

######### Section 7: Surface distance #########
# not done yet.
class SurfaceDistanceImageFilter:
    def __init__(self): self.all_dis = totensor([0])
    def Execute(self, A, B):
        array = sitk.GetArrayViewFromImage
        ADisMap = sitk.Abs(sitk.SignedMaurerDistanceMap(A, squaredDistance = False, useImageSpacing = True))
        BDisMap = sitk.Abs(sitk.SignedMaurerDistanceMap(B, squaredDistance = False, useImageSpacing = True))
        Asurface = sitk.LabelContour(A)
        Bsurface = sitk.LabelContour(B)
        
        # for a pixel 'a' in A, compute aBdis = dis(a, B)
        aBDis = array(BDisMap)[array(Asurface) > 0]
        # for a pixel 'b' in B, compute aBdis = dis(b, A)
        bADis = array(ADisMap)[array(Bsurface) > 0]
        self.all_dis = totensor(np.concatenate((aBDis, bADis), 0))
        
    def GetHausdorffDistance(self): return self.all_dis.max()
    def GetMedianSurfaceDistance(self): return self.all_dis.median()
    def GetAverageSurfaceDistance(self): return self.all_dis.mean()
    def GetDivergenceOfSurfaceDistance(self): return self.all_dis.std()

@torchext
def Metric(A, B, spacing = 1, metric = "HD"):
    '''
    A: (n_batch, n_1, n_2, ..., n_k)
    B: (n_batch, n_1, n_2, ..., n_k)
    return: (n_batch,)
    '''
    A = tonumpy(A) != 0
    B = tonumpy(B) != 0
    spacing = totuple(spacing)
    n_dim = dimof(A) - 1
    n_batch = nbatch(A)
    if len(spacing) == 1: spacing *= n_dim
    Overlap_filter = sitk.LabelOverlapMeasuresImageFilter()
    SD_filter = SurfaceDistanceImageFilter()
    Overlap_execs = {
        'Dice': lambda x: x.GetDiceCoefficient(),
        'Jaccard': lambda x: x.GetJaccardCoefficient(),
        'Volume': lambda x: x.GetVolumeSimilarity(),
        'Falsenegative': lambda x: x.GetFalseNegativeError(),
        'Falsepositive': lambda x: x.GetFalsePositiveError()
    }
    SD_execs = {
        'HD': lambda x: x.GetHausdorffDistance(),
        'MSD': lambda x: x.GetMedianSurfaceDistance(),
        'ASD': lambda x: x.GetAverageSurfaceDistance(),
        'STDSD': lambda x: x.GetDivergenceOfSurfaceDistance()
    }
    measures = np.zeros((n_batch,))
    for b in range(n_batch):
        ITKA = sitk.GetImageFromArray(A[b].astype(np.int), isVector = False)
        ITKA.SetSpacing(spacing)
        ITKB = sitk.GetImageFromArray(B[b].astype(np.int), isVector = False)
        ITKB.SetSpacing(spacing)
        metric = metric.capitalize()
        if metric in Overlap_execs:
            Overlap_filter.Execute(ITKA, ITKB)
            measures[b] = Overlap_execs[metric](Overlap_filter)
        metric = metric.upper()
        if metric in SD_execs:
            SD_filter.Execute(ITKA, ITKB)
            measures[b] = SD_execs[metric](SD_filter)
    return measures

@torchext
def LabelMetric(A, B, spacing = 1, metric = "HD", class_labels = None):
    '''
    :param A: (n_batch, n_1, ..., n_k)
    :param B: (n_batch, n_1, ..., n_k)
    :param class_labels: list[n_class]
    :return: (n_batch, n_class)
    '''
    A = tonumpy(A)
    B = tonumpy(B)
    if not class_labels:
        class_labels = list(set(A.astype(np.int).reshape((-1,))))
        class_labels.sort()
    n_batch = nbatch(A)
    n_class = len(class_labels)
    A_labels = [A == i for i in class_labels]
    B_labels = [B == i for i in class_labels]
    A_maps = np.concatenate(A_labels)
    B_maps = np.concatenate(B_labels)
    metric = Metric(A_maps, B_maps, spacing, metric)
    return metric.reshape((n_class, n_batch)).T

template1 = "@torchext\ndef m{metric}(*args, **kwargs): return Metric(*args, **kwargs, metric = '{metric}')"
template2 = "@torchext\ndef mLabel{metric}(*args, **kwargs): return LabelMetric(*args, **kwargs, metric = '{metric}')"
for metric in ('Dice', 'Jaccard', 'Volume', 'FalseNegative', 'FalsePositive', 'HD', 'MSD', 'ASD', 'STDSD'):
    exec(template1.format(metric=metric))
    exec(template2.format(metric=metric))

###############################################

# Metric abbreviations
metric = dict(
    MI = MutualInformation,
    NMI = NormalizedMutualInformation,
    KL = KLDivergence,
    CLE = CorrelationOfLocalEstimation,
    NVI = NormalizedVectorInformation,
    SSD = SumSquaredDifference,
    MSE = MeanSquaredErrors,
    PSNR = PeakSignaltoNoiseRatio,
    CE = CrossEntropy,
    CC = CrossCorrelation,
    NCC = NormalizedCrossCorrelation,
    SSIM = StructuralSimilarity,
    DSC = LabelDice
)




from pycamia import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = 'pyoverload',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.1.24',
    contact = 'bertiezhou@163.com',
    keywords = ['overload'],
    description = "'pyoverload' overloads the functions by simply using typehints and adding decorator '@overload'. "
).check()

from .typehint import isofsubclass, inheritable, isitertype, iterable, isarray, isdtype, isatype, isoftype, isclassmethod, typename, Type, params, Bool, Int, Float, Str, Set, List, Tuple, Dict, Class, Callable, Function, Method, Lambda, Functional, Real, real, Null, null, Sequence, sequence, Array, Iterable, Scalar, IntScalar, FloatScalar #*
from .override import override, overload, params #*
















































































































from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of reading & writing medical files. ",
    help = "Use `NII` and `DCM` to cope with files.",
    requires = ["nibabel", "pydicom"]
).check()

__all__ = """
    NII
    DCM
""".split()

with __info__:
    import torch
    import batorch as bt
    import nibabel as nib
    import pydicom as dcm
    import numpy as np
    from pycamia import path, shell

def toU(dt):
    if np.dtype(dt).kind == np.dtype(np.int).kind: return np.dtype('uint%d'%(8*dt.itemsize))
    else: return dt
    
def toI(dt):
    if np.dtype(dt).kind == np.dtype(np.uint).kind: return np.dtype('int%d'%(8*dt.itemsize))
    else: return dt

def orderedValue(d):
    return [d[k] for k in sorted(d.keys())]

def orn2quatern(r11, r21, r31, r12, r22, r32):
    eps = 1e-6
    r11, r12, r21, r22 = -r11, -r12, -r21, -r22
    A, B, C = 1, -(r11+r22+2) / 2, (r11+1) * (r22+1) / 4 - ((r12+r21) / 4) ** 2
    quatern_a_sqs = [(-B + np.sqrt(B*B-4*A*C)) / (2*A), (-B - np.sqrt(B*B-4*A*C)) / (2*A)]
    quatern_a = [np.sqrt(np.clip(x, 0, 1)) for x in quatern_a_sqs if -eps <= x <= 1 + eps]
    if len(quatern_a) <= 0: raise TypeError("Invalid orientation")
    if len(quatern_a) > 1:
        r13 = r21 * r32 - r31 * r22
        r23 = - r11 * r32 + r31 * r12
        r33 = r11 * r22 - r21 * r12
        r = np.sqrt(r13 ** 2 + r23 ** 2 + r33 ** 2)
        t = np.abs(r33) / r
        quatern_a = [x for x in quatern_a if np.abs(np.abs(4 * x * x - 1 - r11 - r22) - t) < eps]
        if len(quatern_a) <= 0 or len(set(quatern_a)) > 1: raise TypeError("Invalid orientation")
    qa = np.clip(quatern_a[0], 0, 1)
    if np.abs(qa) > eps:
        qd = (r21 - r12) / qa / 4
        qc = (qd * r32 - qa * r31) / (qd ** 2 + qa ** 2) / 2
        qb = r32 / qa / 2 - qd * qc / qa
    else:
        assert r12 == r21
        qd = np.sqrt(- (r11 + r22) / 2)
        if np.abs(qd) > eps: qb = r31 / qd / 2; qc = r32 / qd / 2
        else: qc = np.sqrt((r22 - r11) / 2); qb = 1 if np.abs(qc) < eps else (r12 / qc / 2)
    return qa, qb, qc, qd
    
def quatern2mat(a, b, c, d):
    return np.array([[a*a+b*b-c*c-d*d, 2*b*c-2*a*d, 2*b*d+2*a*c], 
                    [2*b*c+2*a*d, a*a+c*c-b*b-d*d, 2*c*d-2*a*b], 
                    [2*b*d-2*a*c, 2*c*d+2*a*b, a*a+d*d-c*c-b*b]])
    
def mat2orn(mat): return -mat[0][0], -mat[1][0], mat[2][0], -mat[0][1], -mat[1][1], mat[2][1]
def quatern2orn(*args): return mat2orn(quatern2mat(*args))

def niiheader_to_mat(h):
    R = np.zeros((4, 4))
    R[3][3] = 1
    if h['sform_code'] > 0:
        R[0] = h['srow_x']
        R[1] = h['srow_y']
        R[2] = h['srow_z']
    elif h['qform_code'] == 0:
        R[0][0], R[1][1], R[2][2] = h['pixdim'][1:4]
    else:
        b, c, d = h['quatern_b'], h['quatern_c'], h['quatern_d']
        qx, qy, qz = h['qoffset_x'], h['qoffset_y'], h['qoffset_z']
        qfac, dx, dy, dz = h['pixdim'][:4]
        dz = dz if qfac >= 0 else -dz
        a = np.sqrt(np.clip(1 - (b * b + c * c + d * d), 0, 1))
        R[:3, :3] = quatern2mat(a, b, c, d) * np.array([[dx, dy, dz]])
        R[:3, 3] = qx, qy, qz
    R[:, 0], R[:, 1] = R[:, 1].copy(), R[:, 0].copy()
    return R

def create_nii(dcm, creation):
    data = bt.Tensor(dcm)
    if not isinstance(dcm, DCM) and hasattr(dcm, 'bundle'): bundle = dcm.bundle
    else:
        header = nib.Nifti1Header()
        header['regular'] = b'r'
        header['dim'] = [data.ndim, *data.shape] + [1] * (7 - data.ndim)
        bits = len(data.flatten()[0].tobytes()) * 8
        header['bitpix'] = bits
        header.set_data_dtype(data.numpy().dtype)
        if isinstance(dcm, DCM): meta = dcm.bundle
        else: meta = None
        if meta and 'PixelSpacing' in meta: spacing = [float(x) for x in meta.PixelSpacing]
        else: spacing = [1.0, 1.0]
        if meta and 'SliceThickness' in meta: dz = [float(meta.SliceThickness)]
        else: dz = [1.0]
        header['pixdim'] = [1.0] + spacing + dz + [1.0] * (7 - data.ndim)
        header['qform_code'] = 1
        header['xyzt_units'] = 2
        if meta:
            header['qoffset_x'] = -float(meta.ImagePositionPatient[0])
            header['qoffset_y'] = -float(meta.ImagePositionPatient[1])
            header['qoffset_z'] = float(meta.ImagePositionPatient[2])
            qa, qb, qc, qd = orn2quatern(*[float(x) for x in meta.ImageOrientationPatient])
            header['quatern_b'] = qb
            header['quatern_c'] = qc
            header['quatern_d'] = qd
        bundle = nib.Nifti1Image(data, None, header=header)
    instance = creation(data)
    instance.bundle = bundle
    instance.path = getattr(dcm, 'path', 'Unknown')
    return instance

def nii2dcm(nii, creation):
    data = bt.Tensor(nii)
    if not isinstance(nii, NII) and hasattr(nii, 'bundle'): bundle = nii.bundle
    elif hasattr(nii, 'bundle'):
        header = nii.bundle.header
        if data.ndim > 3: raise TypeError("Dicom is unable to store high dimensional data [%dD]."%data.ndim)
        while data.ndim < 3: data = data.unsqueeze(-1)
        # use_meta_size = header.get('use_meta_size', False)
        # if use_meta_size:
        #     if dimof(data) <= 2: tosize = (self.bundle.Rows, self.bundle.Columns)[:dimof(data)]
        #     else: tosize = (self.bundle.Rows, self.bundle.Columns) + data.shape[2:]
        #     if any([s != 1 for s in scaling]):
        #         raise_rescale()
        #         dt = np.dtype(self.dtype)
        #         mode = 'Nearest' if dt.kind == np.dtype(np.int).kind or dt.kind == np.dtype(np.uint).kind else 'Linear'
        #         data = rescale_to(data.astype(np.float32), tosize, mode = mode).astype(data.dtype)
        # else: tosize = data.shape
        b, c, d = header.get('quatern', (0.0, 0.0, 0.0))
        origin = header.get('origin', (0.0, 0.0, 0.0))
        spacing = header.get('spacing', [1.0] * 8)
        modality = header.get('modality', 'CT')
        if 'header' in header:
            if 'quatern' not in header:
                b, c, d = [header['header'].get('quatern_b', 0.0),
                        header['header'].get('quatern_c', 0.0),
                        header['header'].get('quatern_d', 0.0)]
            if 'origin' not in header:
                origin = [header['header'].get('qoffset_x', 0.0),
                        header['header'].get('qoffset_y', 0.0),
                        header['header'].get('qoffset_z', 0.0)]
            if 'spacing' not in header:
                spacing = header['header'].get('pixdim', [1.0] * 8)
        spacing = spacing[1:4]
        from math import sqrt; a = sqrt(1-b*b-c*c-d*d)
        orn = quatern2orn(a, b, c, d)
        # orn = [-x for x in orn]
        origin = [str(-origin[0]), str(-origin[1]), str(origin[2])]
        slice_thickness = header.get('slice_thickness', spacing[2])
        if 'header' not in header:
            if 'quatern' not in header:
                orn = [float(x) for x in self.bundle.ImageOrientationPatient]
            if 'origin' not in header:
                origin = self.bundle.ImagePositionPatient
            if 'spacing' not in header:
                slice_thickness = float(self.bundle.SliceThickness)
                spacing = [float(x) for x in self.bundle.PixelSpacing] + \
                    [header.get('slice_thickness', abs(slice_thickness))]
            if 'Modality' in self.bundle: modality = self.bundle.Modality
        if 'InstanceCreationTime' in self.bundle: ctime = self.bundle.InstanceCreationTime
        if 'SOPInstanceUID' in self.bundle: UID = self.bundle.SOPInstanceUID
        if 'ContentTime' in self.bundle: time = self.bundle.ContentTime
        if 'TriggerTime' in self.bundle: ttime = self.bundle.TriggerTime
        if 'ReconstructionTargetCenterPatient' in self.bundle:  center = self.bundle.ReconstructionTargetCenterPatient
        bits = len(data.flatten()[0].tobytes()) * 8
        traditional_origin = [-float(origin[0]), -float(origin[1]), float(origin[2])]
        if np.abs(orn[2]) > 0: iz = 0
        elif np.abs(orn[5]) > 0: iz = 1
        else: iz = 2
        position = np.dot(quatern2mat(*orn2quatern(*orn)).T, np.array([traditional_origin]).T)[-1, 0]
        bundles = {}
        typical_slice = float('inf'), None
        Nslice = min(data.shape[-1], header.get('max_slice', float('inf')))
        for slice in range(Nslice):
            # sdcm = dcm.filereader.dcmread(self.bundle.filename, stop_before_pixels=True)
            if not header.get('generate_slices', True) and 0 < slice < Nslice - 1: continue
            sdcm = deepcopy(self.bundle)
            if test(lambda:UID): *segs, tail = UID.split('.')
            if 'SOPInstanceUID' in sdcm:
                sdcm.SOPInstanceUID = '.'.join(segs + [str(int(tail) + slice)])
            if 'ReconstructionTargetCenterPatient' in sdcm and not self.slice_only:
                sdcm.ReconstructionTargetCenterPatient = [0.0, 0.0, center[-1] + slice * slice_thickness]
            if 'TablePosition' in sdcm and not self.slice_only:
                sdcm.TablePosition = position + slice * slice_thickness
            if 'InstanceNumber' in sdcm and not self.slice_only:
                sdcm.InstanceNumber = str(slice + 1)
            if 'ImagePositionPatient' in sdcm and not self.slice_only:
                sdcm.ImagePositionPatient = origin[:iz] + [str(float(origin[iz]) + slice * slice_thickness)] + origin[iz+1:]
            if 'SliceLocation' in sdcm and not self.slice_only:
                sdcm.SliceLocation = str(position + slice * slice_thickness)
            if 'SliceThickness' in sdcm:
                sdcm.SliceThickness = str(abs(slice_thickness))
            if 'InStackPositionNumber' in sdcm and not self.slice_only:
                sdcm.InStackPositionNumber = slice + 1
            if 'ImageOrientationPatient' in sdcm:
                sdcm.ImageOrientationPatient = [str(x) for x in orn]
            # if 'InPlanePhaseEncodingDirection' in sdcm:
            #     del sdcm['InPlanePhaseEncodingDirection']
            if 'Modality' in sdcm and modality:
                sdcm.Modality = modality
            if 'PixelSpacing' in sdcm:
                sdcm.PixelSpacing = [str(x) for x in spacing[:2]]
            if 'BitsStored' in sdcm:
                sdcm.BitsStored = bits
            if 'HighBit' in sdcm:
                sdcm.HighBit = bits - 1
            if 'BitsAllocated' in sdcm:
                sdcm.BitsAllocated = bits
            if 'PixelRepresentation' in sdcm:
                sdcm.PixelRepresentation = int(data.dtype.kind == 'u')
            try:
                self.bundle[0x7005, 0x1018]
                try: sdcm[0x7005, 0x1018]
                except: sdcm[0x7005, 0x1018] = self.bundle[0x7005, 0x1018]
                sdcm[0x7005, 0x1018].value = chr(slice + 1).encode() + chr(0).encode()
            except: pass
            if 'LargestImagePixelValue' in sdcm:
                sdcm.LargestImagePixelValue = np.max(data[..., slice])
            if 'PixelData' in sdcm:
                sdcm.PixelData = data[..., slice].tobytes()
                sdcm['PixelData'].VR = 'OB'
            if 'Rows' in sdcm and 'Columns' in sdcm:
                sdcm.Rows, sdcm.Columns = data.shape[:2]
            if float(sdcm.ImagePositionPatient[2]) < typical_slice[0]:
                typical_slice = float(sdcm.ImagePositionPatient[2]), sdcm
            bundles[slice] = sdcm
        return bundles if header.get('generate_slices', True) else typical_slice[1]

class NII(bt.Tensor):

    def __new__(cls, instance):
        if isinstance(instance, str):
            p = path(instance)
            if not p.ext: p = p // 'nii.gz'
            niiBundle = nib.load(str(p))
            data = niiBundle.get_data()
            self = super().__new__(cls, data)
            self.bundle = niiBundle
            self.path = p
            return self.transpose_(1, 0)
        elif hasattr(instance, 'shape'):
            if instance.ndim == 0: return instance
            if isinstance(instance, NII): return instance
            return create_nii(instance, lambda x: super().__new__(cls, x))
        else: raise TypeError(f"Unknown input for NII: {instance}. ")

    def __enter__(self): return self

    def __exit__(self, *args): return False

    def __call__(self, data=None):
        if data is None: return self.bundle
        data.bundle = self.bundle
        data.path = self.path
        return NII(data)

    def _create_bundle(self, data, use_header_size=False, spacing=None):
        data = bt.Tensor(data)
        header = self.bundle.header.copy()
        if spacing is not None: header['pixdim'] = [1.0] + list(spacing) + [1.0] * (7 - data.ndim)
        if use_header_size:
            raise NotImplementedError("It appears that the developers forgot to implement keyword use_header_size! Please contact us to remind us. ")
            # if any([s != 1 for s in scaling]):
            #     raise_rescale()
            #     dt = np.dtype(self.dtype)
            #     mode = 'Nearest' if dt.kind == np.dtype(np.int).kind or dt.kind == np.dtype(np.uint).kind else 'Linear'
            #     data = rescale_to(data.astype(np.float32), header['dim'][1: dimof(data) + 1], mode = mode).astype(data.dtype)
        else: header['dim'] = [data.ndim] + list(data.shape) + [1] * (7 - data.ndim)
        return nib.Nifti1Image(data.transpose(1, 0), None, header)

    def save(self, path): nib.save(self.bundle, str(path))

    def save_as_dcm(self, path): DCM(self).save(str(path))

    def update(self):
        self.bundle = self._create_bundle(self, False)

    def save_data(self, data, path, use_header_size=False):
        nib.save(self._create_bundle(data, use_header_size), str(path))

    def spacing(self): return self.bundle.header['pixdim'][1: self.ndim + 1]

    # def resample(self, new_spacing):
    #     raise_rescale()
    #     spacing = self.spacing()
    #     new_spacing = totuple(new_spacing)
    #     if len(new_spacing) == 1: new_spacing *= len(spacing)
    #     dt = np.dtype(self.dtype)
    #     mode = 'Nearest' if dt.kind == np.dtype(np.int).kind or dt.kind == np.dtype(np.uint).kind else 'Linear'
    #     new_spacing = tonumpy(new_spacing)
    #     new_data = rescale_to(self.astype(np.float32), 
    #         tuple(int(x) for x in (np.array(self.shape) * spacing / new_spacing).round()), mode = mode).astype(self.dtype)
    #     instance = super().__new__(NII, new_data.shape, dtype=self.dtype)
    #     instance[...] = new_data
    #     instance.path = self.path
    #     instance.bundle = self._create_bundle(new_data, spacing = new_spacing)
    #     return instance

    def affine(self):
        return niiheader_to_mat(self.bundle.header)

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        self = args[0]
        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)

        def apply(r):
            r.bundle = self.bundle
            r.path = self.path

        return bt.Tensor.__torch_function_convert_apply__(ret, apply, cls)

class DCM(bt.Tensor):

    def __new__(cls, instance, slice_only=False):
        if isinstance(instance, str):
            p = path(instance)
            if not p.isdir():
                if not slice_only: p = p@path.Folder
            else: slice_only = False
            dcmBundle = dcm.filereader.dcmread(path(__file__)@path.Folder/"template.dcm")
            slice_arrays = {}
            slices = {}
            zs = {}
            readable = False
            direction_down = True
            for p in ([p] if slice_only else p):
                if not p.ext.lower() in ('dcm', 'ima'): continue
                try: image_slice = dcm.filereader.dcmread(p)
                except: continue
                readable = True
                n_slice = int(image_slice.InstanceNumber)
                if 'SeriesNumber' in image_slice: n_series = int(image_slice.SeriesNumber)
                else: n_series = 0
                try: slice_array = image_slice.pixel_array
                except:
                    try:
                        p_dicom = (p@path.Folder//'dicom').mkdir()/p@path.File
                        if not p_dicom.exists():
                            _, stderr = shell(f"dcmdjpeg {p} {p_dicom}")
                        else: stderr = ''
                        if stderr: raise TypeError("Unknown encoding: %s."%p)
                    except: raise TypeError("Unknown encoding: %s."%p)
                    image_slice = dcm.filereader.dcmread(p_dicom)
                    try: slice_array = image_slice.pixel_array
                    except: raise TypeError("Unknown encoding: %s."%p)
                if n_series not in slices:
                    slice_arrays[n_series] = {}
                    slices[n_series] = {}
                    zs[n_series] = {}
                slice_arrays[n_series][n_slice] = slice_array
                slices[n_series][n_slice] = image_slice
                if image_slice.ImageOrientationPatient[2] != 0: iz = 0
                elif image_slice.ImageOrientationPatient[5] != 0: iz = 1
                else: iz = 2
                if 'ImagePositionPatient' in image_slice: z = float(image_slice.ImagePositionPatient[iz])
                elif 'TablePosition' in image_slice: z = image_slice.TablePosition
                elif 'SliceLocation' in image_slice: z = float(image_slice.SliceLocation)
                else: z = 0.
                zs[n_series][n_slice] = z
            if not readable: raise TypeError("Could not create a DICOM object from " + p + ".")
            sorted_series = sorted([(n_series, slices[n_series]) for n_series in slices], key=lambda x: -len(x[1]))
            n_series = sorted_series[0][0]
            possible_series = [s[1] for s in sorted_series if s[0] == n_series]
            if len(possible_series) >= 8: series = possible_series[7]
            elif len(possible_series) >= 3: series = possible_series[2]
            else: series = possible_series[0]
            min_slice = 1000, None
            max_slice = 0, None
            top_slices = -float('inf'), {}
            bottom_slices = float('inf'), {}
            for n_slice in series:
                image_slice = series[n_slice]
                z = zs[n_series][n_slice]
                if n_slice < min_slice[0]:
                    min_slice = n_slice, image_slice
                if n_slice > max_slice[0]:
                    max_slice = n_slice, image_slice
                if z > top_slices[0]:
                    top_slices = z, {n_slice: image_slice}
                if z < bottom_slices[0]:
                    bottom_slices = z, {n_slice: image_slice}
                if z == top_slices[0]:
                    top_slices[1][n_slice] = image_slice
                if z == bottom_slices[0]:
                    bottom_slices[1][n_slice] = image_slice
            N = min(len(top_slices[1].keys()), len(bottom_slices[1].keys()))
            if N >= 8: i_series = 7
            elif N >= 3: i_series = 2
            else: i_series = 0
            bound1 = sorted(top_slices[1].keys())[i_series]
            bound2 = sorted(bottom_slices[1].keys())[i_series]
            if bound1 > bound2:
                zs = {k: v for k, v in zs[n_series].items() if bound2 <= k <= bound1}
                slices = {k: v for k, v in slice_arrays[n_series].items() if bound2 <= k <= bound1}
                max_slice = bound1, top_slices[1][bound1]
                min_slice = bound2, bottom_slices[1][bound2]
            elif bound1 < bound2:
                zs = {k: v for k, v in zs[n_series].items() if bound1 <= k <= bound2}
                slices = {k: v for k, v in slice_arrays[n_series].items() if bound1 <= k <= bound2}
                max_slice = bound2, bottom_slices[1][bound2]
                min_slice = bound1, top_slices[1][bound1]
            else:
                zs = {k: v for k, v in zs[n_series].items()}
                slices = {k: v for k, v in slice_arrays[n_series].items()}
                bound = sorted(series.keys())[0]
                max_slice = min_slice = bound, series[bound]
            direction_down = zs[max_slice[0]] < zs[min_slice[0]]
            typical_slice = max_slice[1] if direction_down else min_slice[1]
            for key in dir(typical_slice):
                if key == 'PixelData' or '_' in key: continue
                if key.capitalize() != key[0] + key[1:].lower(): continue
                dcmBundle[key] = typical_slice[key]
            ozs = bt.Tensor(sorted(zs.values()))
            if len(set(ozs)) > 1:
                volume = bt.stack(orderedValue({zs[i]: slices[i] for i in slices}), -1)
                dcmBundle.SliceThickness = str(bt.abs(bt.mean(ozs[1:] - ozs[:-1])).item())
            else:
                volume = bt.stack(orderedValue(slices), -1)
            volume = volume.astype(toU(volume.dtype) if dcmBundle.PixelRepresentation else toI(volume.dtype))
            dcmBundle.PixelData = volume.tobytes()
            self = super().__new__(cls, volume)
            self.bundle = dcmBundle
            self.path = path
            self.slice_only = slice_only
            self.update()
            return self
        elif hasattr(instance, 'shape'):
            if instance.ndim == 0: return instance
            if isinstance(instance, DCM): return instance
            if isinstance(instance, NII):
                input = nii2dcmBundle(instance)
            else:
                data = bt.Tensor(instance)
                input.path = 'Unknown'
                self.slice_only = False
            self.update()
            return self
        else: raise TypeError(f"Unknown input for DCM: {instance}. ")

    def __enter__(self): return self
    def __exit__(self, *args): return False
    def __call__(self, data=None):
        if data is None: return self.bundle
        data.bundle = self.bundle
        data.path = self.path
        return DCM(data)
    def _create_bundles(self, data, **header):
        '''
        header: a dict containing 'quatern', 'origin', 'spacing',
            'slice_thickness', 'use_meta_size', 'max_slice',
            'generate_slices', 'modality', 'header'
        '''
        if 'header' not in header and isinstance(data, NII):
            header['header'] = data().header
        data = np.array(data)
        # data = data.transpose(1, 0, *range(2, dimof(data)))
        if dimof(data) > 3: raise TypeError("Dicom is unable to store high dimensional data [%dD]."%dimof(data))
        while dimof(data) < 3: data = np.expand_dims(data, -1)
        use_meta_size = header.get('use_meta_size', False)
        if use_meta_size:
            if dimof(data) <= 2: tosize = (self.bundle.Rows, self.bundle.Columns)[:dimof(data)]
            else: tosize = (self.bundle.Rows, self.bundle.Columns) + data.shape[2:]
            if any([s != 1 for s in scaling]):
                raise_rescale()
                dt = np.dtype(self.dtype)
                mode = 'Nearest' if dt.kind == np.dtype(np.int).kind or dt.kind == np.dtype(np.uint).kind else 'Linear'
                data = rescale_to(data.astype(np.float32), tosize, mode = mode).astype(data.dtype)
        else: tosize = data.shape
        b, c, d = header.get('quatern', (0.0, 0.0, 0.0))
        origin = header.get('origin', (0.0, 0.0, 0.0))
        spacing = header.get('spacing', [1.0] * 8)
        modality = header.get('modality', 'CT')
        if 'header' in header:
            if 'quatern' not in header:
                b, c, d = [header['header'].get('quatern_b', 0.0),
                        header['header'].get('quatern_c', 0.0),
                        header['header'].get('quatern_d', 0.0)]
            if 'origin' not in header:
                origin = [header['header'].get('qoffset_x', 0.0),
                        header['header'].get('qoffset_y', 0.0),
                        header['header'].get('qoffset_z', 0.0)]
            if 'spacing' not in header:
                spacing = header['header'].get('pixdim', [1.0] * 8)
        spacing = spacing[1:4]
        from math import sqrt; a = sqrt(1-b*b-c*c-d*d)
        orn = quatern2orn(a, b, c, d)
        # orn = [-x for x in orn]
        origin = [str(-origin[0]), str(-origin[1]), str(origin[2])]
        slice_thickness = header.get('slice_thickness', spacing[2])
        if 'header' not in header:
            if 'quatern' not in header:
                orn = [float(x) for x in self.bundle.ImageOrientationPatient]
            if 'origin' not in header:
                origin = self.bundle.ImagePositionPatient
            if 'spacing' not in header:
                slice_thickness = float(self.bundle.SliceThickness)
                spacing = [float(x) for x in self.bundle.PixelSpacing] + \
                    [header.get('slice_thickness', abs(slice_thickness))]
            if 'Modality' in self.bundle: modality = self.bundle.Modality
        if 'InstanceCreationTime' in self.bundle: ctime = self.bundle.InstanceCreationTime
        if 'SOPInstanceUID' in self.bundle: UID = self.bundle.SOPInstanceUID
        if 'ContentTime' in self.bundle: time = self.bundle.ContentTime
        if 'TriggerTime' in self.bundle: ttime = self.bundle.TriggerTime
        if 'ReconstructionTargetCenterPatient' in self.bundle:  center = self.bundle.ReconstructionTargetCenterPatient
        bits = len(data.flatten()[0].tobytes()) * 8
        traditional_origin = [-float(origin[0]), -float(origin[1]), float(origin[2])]
        if np.abs(orn[2]) > 0: iz = 0
        elif np.abs(orn[5]) > 0: iz = 1
        else: iz = 2
        position = np.dot(quatern2mat(*orn2quatern(*orn)).T, np.array([traditional_origin]).T)[-1, 0]
        bundles = {}
        typical_slice = float('inf'), None
        Nslice = min(data.shape[-1], header.get('max_slice', float('inf')))
        for slice in range(Nslice):
            # sdcm = dcm.filereader.dcmread(self.bundle.filename, stop_before_pixels=True)
            if not header.get('generate_slices', True) and 0 < slice < Nslice - 1: continue
            sdcm = deepcopy(self.bundle)
            if test(lambda:UID): *segs, tail = UID.split('.')
            if 'SOPInstanceUID' in sdcm:
                sdcm.SOPInstanceUID = '.'.join(segs + [str(int(tail) + slice)])
            if 'ReconstructionTargetCenterPatient' in sdcm and not self.slice_only:
                sdcm.ReconstructionTargetCenterPatient = [0.0, 0.0, center[-1] + slice * slice_thickness]
            if 'TablePosition' in sdcm and not self.slice_only:
                sdcm.TablePosition = position + slice * slice_thickness
            if 'InstanceNumber' in sdcm and not self.slice_only:
                sdcm.InstanceNumber = str(slice + 1)
            if 'ImagePositionPatient' in sdcm and not self.slice_only:
                sdcm.ImagePositionPatient = origin[:iz] + [str(float(origin[iz]) + slice * slice_thickness)] + origin[iz+1:]
            if 'SliceLocation' in sdcm and not self.slice_only:
                sdcm.SliceLocation = str(position + slice * slice_thickness)
            if 'SliceThickness' in sdcm:
                sdcm.SliceThickness = str(abs(slice_thickness))
            if 'InStackPositionNumber' in sdcm and not self.slice_only:
                sdcm.InStackPositionNumber = slice + 1
            if 'ImageOrientationPatient' in sdcm:
                sdcm.ImageOrientationPatient = [str(x) for x in orn]
            # if 'InPlanePhaseEncodingDirection' in sdcm:
            #     del sdcm['InPlanePhaseEncodingDirection']
            if 'Modality' in sdcm and modality:
                sdcm.Modality = modality
            if 'PixelSpacing' in sdcm:
                sdcm.PixelSpacing = [str(x) for x in spacing[:2]]
            if 'BitsStored' in sdcm:
                sdcm.BitsStored = bits
            if 'HighBit' in sdcm:
                sdcm.HighBit = bits - 1
            if 'BitsAllocated' in sdcm:
                sdcm.BitsAllocated = bits
            if 'PixelRepresentation' in sdcm:
                sdcm.PixelRepresentation = int(data.dtype.kind == 'u')
            try:
                self.bundle[0x7005, 0x1018]
                try: sdcm[0x7005, 0x1018]
                except: sdcm[0x7005, 0x1018] = self.bundle[0x7005, 0x1018]
                sdcm[0x7005, 0x1018].value = chr(slice + 1).encode() + chr(0).encode()
            except: pass
            if 'LargestImagePixelValue' in sdcm:
                sdcm.LargestImagePixelValue = np.max(data[..., slice])
            if 'PixelData' in sdcm:
                sdcm.PixelData = data[..., slice].tobytes()
                sdcm['PixelData'].VR = 'OB'
            if 'Rows' in sdcm and 'Columns' in sdcm:
                sdcm.Rows, sdcm.Columns = data.shape[:2]
            if float(sdcm.ImagePositionPatient[2]) < typical_slice[0]:
                typical_slice = float(sdcm.ImagePositionPatient[2]), sdcm
            bundles[slice] = sdcm
        return bundles if header.get('generate_slices', True) else typical_slice[1]
    def save(self, path): self.save_data(self, str(path))
    def save_as_nii(self, path): NII(self).save(str(path))
    def update(self, header=None):
        if not isinstance(self.bundle, dcm.dataset.Dataset):
            self.bundle = dcm.filereader.dcmread(Template)
        if header: self.bundle = self._create_bundles(self, generate_slices=False, header=header)
        else: self.bundle = self._create_bundles(self, generate_slices=False)
    def save_data(self, data, fpath, **header):
        fpath = str(fpath)
        if os.path.isfile(fpath):
            path = os.path.dirname(fpath)
            pmkdir(path)
        else:
            pmkdir(fpath)
            if '.' in os.path.basename(fpath):
                path = os.path.dirname(fpath)
            else:
                path = fpath
                fpath = os.path.join(path, "slice.dcm")
        slices = self._create_bundles(data, **header)
        if self.slice_only: slices[0].save_as(fpath); return
        for i, s in slices.items():
            *pfix, sfix = fpath.split('.')
            s.save_as('.'.join(pfix) + "_%04d"%(i + 1) + '.' + sfix)
    def spacing(self):
        meta = self()
        if 'PixelSpacing' in meta: spacing = [float(x) for x in meta.PixelSpacing]
        else: spacing = [1.0, 1.0]
        if 'SliceThickness' in meta: dz = [float(meta.SliceThickness)]
        else: dz = [1.0]
        spacing = spacing + dz
        return np.array(spacing)
    def resample(self, new_spacing):
        raise_rescale()
        spacing = self.spacing()
        new_spacing = totuple(new_spacing)
        if len(new_spacing) == 1: new_spacing *= len(spacing)
        dt = np.dtype(self.dtype)
        mode = 'Nearest' if dt.kind == np.dtype(np.int).kind or dt.kind == np.dtype(np.uint).kind else 'Linear'
        new_spacing = tonumpy(new_spacing)
        new_data = rescale_to(self.astype(np.float32), 
            tuple(int(x) for x in (np.array(self.shape) * spacing / new_spacing).round()), mode = mode).astype(self.dtype)
        instance = super().__new__(DCM, new_data.shape, dtype=self.dtype)
        instance[...] = new_data
        instance.path = self.path
        instance.slice_only = self.slice_only
        instance.bundle = self._create_bundle(new_data, spacing = new_spacing)
        return instance
    def affine(self):
        return niiheader_to_mat(NII(self)().header)



__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to manage package info.",
    help = "Use `info_manager(**kwargs)` to create an info object and use `with info` to check for imports. "
)

__all__ = """
    info_manager
""".split()

import os, re

from .environment import get_environ_globals
from .strop import tokenize

class info_manager:
    """
    info_manager() -> info_object

    An object storing the info of the current file.

    Note:
        It is currently provided for private use in project PyCAMIA only. 
        But it can also be used outside.

    Example:
    ----------
    Code:
        __info__ = info_manager(project="PyCAMIA", package="", requires="xxx").check()
        print("Project is", __info__.project)
        with __info__:
            import xxx
            from xxx import yyy
    Output:
        Project is PyCAMIA
    Error if xxx doesnot exist: ModuleNotFoundError
    Error if yyy doesnot exist in xxx: ImportError
    """
    
    @staticmethod
    def parse(code):
        info = eval(code)
        raw_args = tokenize(code, sep=['(',')'])[1]
        info.tab = raw_args[:re.search(r'\w', raw_args).span()[0]].strip('\n')
        info.order = tokenize(raw_args, sep=[',', '='], strip=None)[::2]
        return info

    def __init__(self, project = "", package = "", requires = "", **properties):
        if isinstance(requires, str): requires = requires.split()
        properties.update(dict(project=project, package=package, requires=requires))
        self.properties = properties
        self.__dict__.update(properties)
        file_path = get_environ_globals()['__file__']
        file = os.path.extsep.join(os.path.basename(file_path).split(os.path.extsep)[:-1])
        self.name = '.'.join([x for x in [project, package, file] if x and x != "__init__"])
        self.tab = ' ' * 4
        major_keys = "project package requires".split()
        self.order = major_keys + list(set(properties.keys()) - set(major_keys))
        
    def check_requires(self):
        not_found_packages = []
        for r in self.requires:
            tokens = tokenize(r, ['<', '>', '='], strip=None, keep_empty=False)
            if len(tokens) == 2: rname, rversion = tokens
            else: rname, rversion = tokens[0], None
            try:
                package = __import__(rname)
                if rversion is not None:
                    op = r.replace(rname, '').replace(rversion, '').strip()
                    if not eval(repr(tuple(float(x) for x in package.__version__.split('.'))) + op + repr(tuple(float(x) for x in rversion.split('.')))):
                        not_found_packages.append(r)
            except ModuleNotFoundError: not_found_packages.append(rname)
        if len(not_found_packages) > 0:
            raise ModuleNotFoundError(f"'{self.name}' cannot be used without dependencies {repr(not_found_packages)}.")
            
    def check(self):
        self.check_requires()
        return self
    
    def version_update(self):
        if hasattr(self, 'version'):
            version = re.sub(r"((\d+\.){2})(\d+)", lambda x: x.group(1)+str(eval(x.group(3))+1), self.version)
        else: version = '1.0.0'
        self.version = version
    
    def __enter__(self): return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type == ModuleNotFoundError:
            module_name = str(exc_value).split("'")[1]
            raise ModuleNotFoundError(f"'{self.name}' cannot be used without dependency '{module_name}'. ")
        elif exc_type == ImportError:
            func_name, _, module_name = str(exc_value).split("'")[1:4]
            raise ImportError(f"'{self.name}' requires '{func_name}' in module '{module_name}'. Please double check the version of packages. ")
        
    def __str__(self):
        args = ',\n'.join([self.tab + k + ' = ' + repr(getattr(self, k)) for k in self.order])
        return "info_manager(\n" + args + "\n)"
    
    def __getitem__(self, name):
        return getattr(self, name)
    
    def __setattr__(self, name, value):
        if hasattr(self, 'order') and name not in self.order: self.order.append(name)
        super().__setattr__(name, value)
    
    def get(self, name, value):
        if hasattr(self, name): return self[name]
        return value




from pycamia import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = 'batorch',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.0.25',
    contact = 'bertiezhou@163.com',
    keywords = ['torch', 'batch', 'batched data'],
    description = "'batorch' is an extension of package torch, for tensors with batch dimensions. ",
    requires = ['torch']
).check()

import torch
distributed = torch.distributed
autograd = torch.autograd
optim = torch.optim
utils = torch.utils
nn = torch.nn















































































































from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to visualize features of 3D volume.",
    help = "Use `from micomputing import plot as plt`.",
    requires = "matplotlib"
).check()

__all__ = """
    background
    imshow
    maskshow
    bordershow
""".split()

with __info__:
    import batorch as bt
    from matplotlib import pyplot as plt
    from matplotlib import colors as mc
    from matplotlib.pyplot import *
    from pyoverload import *
    from pycamia import prod
    __all__.extend([x for x in dir(plt) if not x.startswith('_') and x not in __all__])

def totuple(x, l=None):
    if not isinstance(x, tuple): x = (x,)
    if l is None: l = len(x)
    if l % len(x) == 0: return x * (l // len(x))
    raise TypeError(f"{x} can not be converted into a tuple of length {l}. ")

canvas = None
colors = ['red', 'green', 'blue', 'gold', 'purple', 'gray', 'pink', 'darkgreen', 'dodgerblue']
assert all(c in mc.CSS4_COLORS for c in colors)

def to_image(data: Array, nslice: [int, null]=None, dim: int=-1, has_cmap=False):
    data = bt.tensor(data).squeeze()
    if data.ndim <= 1: raise TypeError("Please don't use 'plot.imshow' to demonstrate an array or a scalar. ")
    if data.nspace > 3: raise TypeError(f"'plot.imshow' takes 2 or 3D-data as input (currently {data.shape}), please reduce the dimension manually or specify special dimensions to reduce. ")
    if data.nspace == 3:
        if data.has_batch: data = data.sample(random=False, dim=[])
        if data.has_channel: data = data.sample(random=False, dim={})
        if nslice is None:
            if data.space[-1] <= 3: pass
            elif data.space[0] <= 3: data = data.mvdim(0, 2)
            else:
                nslice = data.space[-1] // 2
                data = data.pick(nslice, dim)
        else: data = data.pick(nslice, dim)
    elif data.nspace == 2:
        if data.has_batch: data = data.sample(random=False, dim=[])
        if data.has_channel:
            if has_cmap: data = data.sample(random=False, dim={})
            else: data = data.sample(number=min(data.channel_size, 3), random=False, dim={}).mvdim(data.channel_dimension, -1)
    elif data.ndim == 3: data = data.sample(random=False, dim=[])
    return data.float().normalize()

def to_RGB(*color):
    if len(color) == 0: return (1.,) * 3
    elif len(color) == 1:
        c = color[0]
        if isinstance(c, float) and 0 <= c <= 1: return (c,) * 3
        elif isinstance(c, (int, float)) and 0 <= c <= 255: return (c / 255,) * 3
        elif isinstance(c, tuple): return to_RGB(*c)
        elif isinstance(c, str):
            if not c.startswith('#'): c = mc.BASE_COLORS.get(c.lower(), mc.CSS4_COLORS.get(c.lower(), '#FFF'))
            if isinstance(c, tuple): return to_RGB(*c)
            return mc.hex2color(c)
        else: raise TypeError("Unaccepted color type. ")
    elif len(color) == 3:
        if all(isinstance(c, float) and 0 <= c <= 1 for c in color): return color
        elif all(isinstance(c, (int, float)) and 0 <= c <= 255 for c in color): return tuple(c / 255 for c in color)
        else: raise TypeError("Unaccepted color type. ")
    else: raise TypeError("Unaccepted color type. ")


@params
def imshow(data: [Array, null]=None, nslice: [int, null]=None, dim: int=-1, **kwargs):
    """
    An automatic image display function for all kinds of tensors. 
    The first image in batched images will be selected to be showed. 
    For medical images:
    Displacements with channel dimension identified will be displayed as RGB colored maps.
    If there are no dimension <=3, gray scaled images will be showed. 
    Transverse medical image with the right hand side of the subject shown on the left
        and anterior shown at the bottom will be selected for 3D volumes.
    `nslice` and `dim` are used for 3D volumes only, meaning to show the `nslice` slice of dimension `dim`. 
    """
    global canvas
    has_cmap = True
    if 'cmap' not in kwargs:
        has_cmap = False
        kwargs['cmap'] = plt.cm.gray
    if data is not None: canvas = to_image(data, nslice, dim, has_cmap)
    if canvas is None or isinstance(canvas, tuple):
        raise TypeError("Please input data in 'imshow' or 'background' to show. ")
    return plt.imshow(canvas.numpy(), **kwargs)

def background(*color):
    """
    Set a background color by RGB or a gray scale, conflict with imshow. 
    """
    global canvas
    canvas = to_RGB(*color)

@params
def maskshow(*masks, on=None, alpha=0.5, nslice=None, dim=-1, stretch=False, **kwargs):
    global canvas
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isarray(on): canvas = to_image(on, nslice, dim)
        elif isinstance(on, list): canvas = to_image(Tensor(on), nslice, dim)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
    if len(masks) == 0: return imshow
    alpha = totuple(alpha, len(masks))
    new_masks = []
    new_alpha = []
    for m, a in zip(masks, alpha):
        img = to_image(m, nslice, dim)
        if img.ndim == 3:
            new_masks.extend(x.squeeze(-1) for x in img.split(1, dim=dim))
            new_alpha.extend([a] * img.size(dim))
        else:
            new_masks.append(img)
            new_alpha.append(a)
    color_mask_map = [(to_RGB(c), m, a) for c, m, a in zip(colors*(len(new_masks) // len(colors) + 1), new_masks, new_alpha)]
    color_mask_map.extend((to_RGB(c), m, alpha[0]) for c, m in kwargs.items())

    if not stretch:
        shapes = [m.ishape for _, m, _ in color_mask_map]
        target_shape = shapes[0]
        if len(set(shapes)) > 1 or not isinstance(canvas, tuple) and target_shape != canvas.shape:
            raise TypeError("Please use masks of the same size as the background image, "
                            "or use 'stretch=True' in 'maskshow' to automatically adjust the image sizes. ")
    else:
        def adjust(m, to):
            ms = tuple(m.shape)
            scaling = tuple((a // b, b // a) for a, b in zip(to, ms))
            return m.down_scale([max(v, 1) for u, v in scaling]).up_scale([max(u, 1) for u, v in scaling]).crop_as(to)
        shapes = [m.ishape for _, m, _ in color_mask_map]
        if not isinstance(canvas, tuple): shapes.append(canvas.shape[:2])
        areas = [u * v for u, v in shapes]
        target_shape = shapes[areas.index(max(areas))]
        color_mask_map = [(c, adjust(m, to=target_shape), a) for c, m, a in color_mask_map]
        canvas = adjust(canvas, to=target_shape)

    target_shape = bt.Size(*target_shape, {3})
    if isinstance(canvas, tuple): canvas = bt.tensor(list(canvas)).expand_to(target_shape)
    elif canvas.ndim == 2: canvas = canvas.expand_to(target_shape)
    coeff = prod(1 - a * m for _, m, a in color_mask_map)
    canvas *= coeff
    for i, (c, m, a) in enumerate(color_mask_map):
        coeff = prod(a * m if j == i else 1 - a * m for j, (_, m, a) in enumerate(color_mask_map))
        canvas += coeff.unsqueeze(-1) * m.unsqueeze(-1) * bt.tensor(list(c)).unsqueeze(0, 1)

    return plt.imshow(canvas.numpy(), **kwargs)

def smooth(curve):
    """
    curve: 2xn
    """
    middle = (curve[:, :-2] + curve[:, 1:-1] + curve[:, 2:]) / 3
    if all(curve[:, 0] == curve[:, -1]):
        head = (curve[:, :1] + curve[:, 1:2] + curve[:, -2:-1]) / 3
        return bt.cat(head, middle, head, dim=1)
    return bt.cat(curve[:, :1], middle, curve[:, -1:], dim=1)

def sharpen(curve, old_curve):
    """
    sharpen towards old_curve
    curve, old_curve: 2xn
    """
    a = bt.cat(curve[:, -2:-1] - curve[:, -1:], curve[:, :-1] - curve[:, 1:], dim=1)
    b = bt.cat(curve[:, 1:] - curve[:, :-1], curve[:, 1:2] - curve[:, :1], dim=1)
    costheta = (a*b).sum(0) / bt.sqrt((a*a).sum(0)) / bt.sqrt((b*b).sum(0))
    new_curve = curve.clone()
    new_curve[:, costheta > -0.9] = old_curve[:, costheta > -0.9]
    return new_curve

def constraint(new_curve, old_curve, constraint_curve):
    """
    constraint new_curve towards constraint_curve
    curve, old_curve: 2xn
    """
    dis_sqs = bt.sum((new_curve - constraint_curve) ** 2, 0)
    percentile = 2 #min(np.sort(dis_sqs)[98 * len(dis_sqs) // 100], 1)
    return bt.where(dis_sqs <= percentile, new_curve, old_curve)

def border(mask, min_length = 10):
    grid = bt.image_grid(*mask.shape)
    mask = mask > 0.5
    idx = mask[1:, :] ^ mask[:-1, :]
    idx = idx.expand_to(2, -1, mask.size(1))
    locs1 = (grid[:, 1:, :] + grid[:, :-1, :])[idx] / 2
    idx = mask[:, 1:] ^ mask[:, :-1]
    idx = idx.expand_to(2, mask.size(0), -1)
    locs2 = (grid[:, :, 1:] + grid[:, :, :-1])[idx] / 2
    locs = bt.cat(locs1.reshape(2, -1), locs2.reshape(2, -1), dim=1)
    if locs.size == 0: return []
    curves = []
    unvisited = bt.ones(locs.shape[-1])
    while True:
        if not any(unvisited): break
        first = bt.argmax(unvisited).item()
        cloc = locs[:, first:first + 1]
        unvisited[first] = 0
        curve = cloc
        while True:
            dissq = bt.sum((locs - cloc) ** 2, 0)
            inloc = bt.argmax(bt.where((unvisited > 0) & (dissq > 0), 1/dissq.clamp(min=0.1), bt.tensor(0).float()))
            if dissq[inloc] > 2: break
            cloc = locs[:, inloc:inloc + 1]
            curve = bt.cat(curve, cloc, dim=1)
            unvisited[inloc] = 0
            if not any(unvisited): break
        sloc = locs[:, first:first + 1]
        if bt.sum((cloc - sloc) ** 2) <= 2:
            curve = bt.cat(curve, sloc, dim=1)
        if curve.shape[1] <= min_length: continue
        scurve = curve
        for _ in range(100): scurve = constraint(smooth(scurve), scurve, curve)
        ccurve = scurve
        for _ in range(100):
            scurve = constraint(sharpen(scurve, curve), scurve, ccurve)
            scurve = constraint(smooth(scurve), scurve, curve)
        curves.append(scurve)
    return curves

def bordershow(*masks, on=None, mask_alpha=0., nslice=None, dim=-1, stretch=False, min_length = 10, **kwargs):
    global canvas
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isarray(on): canvas = to_image(on, nslice, dim)
        elif isinstance(on, list): canvas = to_image(Tensor(on), nslice, dim)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
    if len(masks) == 0: return
    new_masks = []
    for m in masks:
        img = to_image(m, nslice, dim)
        if img.ndim == 3:
            new_masks.extend(x.squeeze(-1) for x in img.split(1, dim=dim))
        else:
            new_masks.append(img)
    color_mask_map = [(to_RGB(c), m) for c, m in zip(colors*(len(new_masks) // len(colors) + 1), new_masks)]
    color_mask_map.extend((to_RGB(c), m) for c, m in kwargs.items())

    new_masks = [m for _, m in color_mask_map]
    shapes = [m.ishape for _, m in color_mask_map]
    if not stretch:
        target_shape = shapes[0]
        if len(set(shapes)) > 1 or not isinstance(canvas, tuple) and target_shape != canvas.shape:
            raise TypeError("Please use masks of the same size as the background image, "
                            "or use 'stretch=True' in 'maskshow' to automatically adjust the image sizes. ")
    else:
        def adjust(m, to):
            ms = tuple(m.shape)
            scaling = tuple((a // b, b // a) for a, b in zip(to, ms))
            return m.down_scale([max(v, 1) for u, v in scaling]).up_scale([max(u, 1) for u, v in scaling]).crop_as(to)
        if not isinstance(canvas, tuple): shapes.append(canvas.shape[:2])
        areas = [u * v for u, v in shapes]
        target_shape = shapes[areas.index(max(areas))]
        color_mask_map = [(c, adjust(m, to=target_shape)) for c, m in color_mask_map]
        canvas = adjust(canvas, to=target_shape)

    if mask_alpha > 0: maskshow(*new_masks, alpha=mask_alpha, **kwargs)
    else: imshow(**kwargs)
    plots = []
    for color, mask in color_mask_map:
        curves = border(mask, min_length)
        for c in curves:
            plots.append(plt.plot(c[1], c[0], color = color, **kwargs))
    return plots





from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to handle the exceptions.",
    help = "Handle the exceptions. `touch` swallow the error, `crashed` to see if there is an erro, `avouch` to assert with text, `Error` to create new error types. "
)

__all__ = """
    touch
    crashed
    avouch
    Error
""".split()

import re, sys
from typing import Callable, Union

from .environment import get_args_expression, get_environ_locals, get_environ_globals
from .functions import const_function

def touch(v: Union[Callable, str], default=None):
    """
    Touch a function or an expression `v`, see if it causes exception. 
    If not, output the result, otherwise, output `default`. 
    
    Note:
        Use `default = pycamia.functions.identity_function` (or write one yourself)
        to return the exception object.
    
    Example:
    ----------
    >>> a = 0
    >>> touch(lambda: 1/a, default = 'fail')
    fail
    """
    if not callable(default):
        default = const_function(default)
    if isinstance(v, str):
        local_vars = get_environ_locals()
        local_vars.update(locals())
        locals().update(local_vars)
        try: return eval(v)
        except Exception as e: return default(e)
    else:
        try: return v()
        except Exception as e: return default(e)

def avouch(v: bool, txt=""):
    """
    Assert with text. 
    
    Inputs:
        v[bool]: the expression to be validated.
        txt[str]: the assertion message when the test fails.
    """
    if not v:
        if not txt:
            expr = get_args_expression().split(',')[0].strip()
            if (expr is not None):
                txt = f"Failure in assertion '{expr}'"
        raise AssertionError(txt)

def crashed(func):
    """
    Validate whether a function `func` would crash. 
    """
    try:
        func()
    except:
        return True
    return False

def Error(name: str):
    """
    Create a temporary error by text. 

    Inputs:
        name[str]: the name of the error; It is used to identify the error type. 

    Example:
    ----------
    >>> try:
    >>>     raise Error("TEST")()
    >>> except Error("TEST"):
    >>>     print('caught')
    ... 
    caught
    """
    v = get_environ_globals()
    error_name = f"{name}Error"
    if error_name in v: return v[error_name]
    exec(f"class {error_name}(Exception): pass")
    v[error_name] = eval(error_name)
    return eval(error_name)



#! python3.8 -u
#  -*- coding: utf-8 -*-

##############################
## Project PyCTLib
## Package torchplus
##############################

__all__ = """
    Device
    DeviceCPU
    Tensor
    Size
    set_autodevice
    unset_autodevice
    is_autodevice
""".split()

try:
    import torch
except ImportError:
    raise ImportError("'pyctlib.torchplus' cannot be used without dependency 'torch'.")
import typing
import inspect
import builtins
import torchplus as tp
from .tensorfunc import __all__ as tf_list
# from pyoverload import overload, override, Tuple, List, Set, params, null, Array, isarray, isoftype, isofsubclass, isint, isdtype, isitertype, isclassmethod
from pyoverload import *
from pyctlib import raw_function, return_type_wrapper, touch, vector
#from pyctlib.visual.debugger import profile
from functools import wraps
from types import GeneratorType, MethodWrapperType
from collections import OrderedDict
from .torch_namespace import *
from .device import AutoDevice as Device, DeviceCPU

"""
from torchplus import Tensor
import torchplus as tp
"""

# Device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

_auto_device = True

def set_autodevice(flag=True):
    global _auto_device
    _auto_device = flag

def unset_autodevice():
    global _auto_device
    _auto_device = False

def is_autodevice():
    global _auto_device
    return _auto_device

INT = builtins.int
MIN = builtins.min
MAX = builtins.max
ANY = builtins.any
ALL = builtins.all
RANGE = builtins.range
FLOAT = builtins.float
NUM = (INT, FLOAT)

def kwget(kwargs, key, default):
    if kwargs is None: return default
    else: return kwargs.get(key, default)

class Size(tuple):

    NegSizeError = TypeError("Size cannot have negative values except -1 indicating arbitrary number. ")

    def __new__(cls, *args, **kwargs):

        if len(args) == 1:
            arg = args[0]
            if hasattr(arg, 'shape'): arg = arg.shape
            if isinstance(arg, (INT, set)): pass
            elif isinstance(arg, tuple): args = arg
            elif isinstance(arg, GeneratorType):
                self = super().__new__(cls, arg)
                self.ndim = len(tuple(arg))
                return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))
            elif isinstance(arg, list):
                l = len(arg)
                if l != 1:
                    self = super().__new__(cls, arg)
                    self.ndim = l
                    return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))
            elif isinstance(arg, Size):
                self = super().__new__(cls, arg.tuple())
                self.ndim = arg.ndim
                self._special = arg._special
                self._batch_first = arg._batch_first
                return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))
            else:
                raise TypeError("'Size' object only takes tensors, tuples, lists and generators as initialization. ")

        # args is now a tuple of * or [*], {*} where * is an integer.
        ndim = len(args)
        batch_dim = ndim
        channel_dim = ndim
        raw_args = []
        for i, a in enumerate(args):

            if isinstance(a, list):
                if batch_dim == ndim:
                    batch_dim = i
                else:
                    raise TypeError("Only one batch dimension is allowed.")
                raw_args.append(a[0])

            elif isinstance(a, set):
                if channel_dim == ndim:
                    channel_dim = i
                else:
                    raise TypeError("Only one channel dimension is allowed.")
                raw_args.append(a.pop())

            else: raw_args.append(a)

        self = super().__new__(cls, raw_args)
        self.ndim = ndim
        return self.set_special_(kwargs.get('batch_dim', batch_dim), kwargs.get('channel_dim', channel_dim))

    def tuple(self): return super().__new__(tuple, self)

    @property
    def batch_dimension(self): return self._special[0 if self._batch_first else 1]

    @batch_dimension.setter
    def batch_dimension(self, batch_dim):
        self.set_special_(batch_dim, None)

    def batch_dimension_(self, value):
        self.batch_dimension = value
        return self

    @property
    def batch_size(self):
        batch_dim = self.batch_dimension
        if batch_dim == self.ndim:
            raise ValueError("There is no batch dimension provided. ")
        return self[batch_dim]

    @property
    def channel_dimension(self): return self._special[1 if self._batch_first else 0]

    @channel_dimension.setter
    def channel_dimension(self, channel_dim):
        self.set_special_(None, channel_dim)

    def channel_dimension_(self, value):
        self.channel_dimension = value
        return self

    @property
    def channel_size(self):
        channel_dim = self.channel_dimension
        if channel_dim is None:
            raise ValueError("There is no channel dimension provided. ")
        return self[channel_dim]

    @property
    def special(self): return [x for x in self._special if x < self.ndim]

    def special_from_(self, other=None):

        if isinstance(other, Size) or isinstance(other, Tensor) and other.init:
            self._special = [self.ndim if x == other.ndim else x for x in other._special]
            self._batch_first = other._batch_first
        else:
            self._special = [self.ndim, self.ndim]
            self._batch_first = True

        return self

    def add_special_from_(self, other=None):

        if isinstance(other, Size):

            doit = False
            batch_dim = None
            channel_dim = None

            if self.batch_dimension == self.ndim:
                batch_dim = other.batch_dimension
                if batch_dim != self.ndim:
                    doit = True
            if self.channel_dimension == self.ndim:
                channel_dim = other.channel_dimension
                if channel_dim != self.ndim:
                    doit = True
            if doit: self.set_special_(batch_dim, channel_dim)

        return self

    def set_special_(self, batch_dim=None, channel_dim=None):

        if batch_dim is not None:
            if not isinstance(batch_dim, INT): batch_dim = self.ndim
            if batch_dim < 0: batch_dim = batch_dim + self.ndim
            if not 0 <= batch_dim <= self.ndim:
                raise TypeError(f"batch_dimension should be a dimension index which is smaller than {self.ndim}. ")
        else: batch_dim = self.ndim

        if channel_dim is not None:
            if not isinstance(channel_dim, INT): channel_dim = self.ndim
            if channel_dim < 0: channel_dim = channel_dim + self.ndim
            if not 0 <= channel_dim <= self.ndim:
                raise TypeError(f"channel_dimension should be a dimension index which is smaller than {self.ndim}. ")
        else: channel_dim = self.ndim

        if batch_dim is None and channel_dim is None: return self

        if batch_dim < channel_dim:
            self._batch_first = True
            self._special = [batch_dim, channel_dim]
        elif channel_dim < batch_dim:
            self._batch_first = False
            self._special = [channel_dim, batch_dim]
        elif batch_dim < self.ndim:
            raise ValueError(f"special dimensions can not be the same: {batch_dim} and {channel_dim}. ")
        else:
            self._batch_first = True
            self._special = [channel_dim, channel_dim]

        return self

    def insert_special_to_tuple(self, target, value):
        s = self._special
        t = tuple(target)
        res = t[:s[0]]
        if s[0] < self.ndim: res += (value,)
        else: return Size(res).special_from_(self)
        res += t[s[0]:s[1]-1]
        if s[1] < self.ndim: res += (value,)
        else: return Size(res).special_from_(self)
        res += t[s[1]-1:]
        return Size(res).special_from_(self)

    def replace_special(self, value):
        s = self._special
        t = tuple(self)
        res = t[:s[0]]
        if s[0] < self.ndim: res += (value,)
        else: return Size(res).special_from_(self)
        res += t[s[0]+1:s[1]]
        if s[1] < self.ndim: res += (value,)
        else: return Size(res).special_from_(self)
        res += t[s[1]+1:]
        return Size(res).special_from_(self)

    @property
    def space(self):
        s = self._special
        t = tuple(self)
        return t[:s[0]] + t[s[0]+1:s[1]] + t[s[1]+1:]

    @property
    def nele(self):
        p = 1
        for i in self:
            if i == -1: return -1
            p *= i
        return p

    def __len__(self): return self.ndim

    @property
    def n_dim(self): return self.ndim

    @property
    def nspace(self): return self.ndim - (self._special[0] < self.ndim) - (self._special[1] < self.ndim)

    @property
    def nspecial(self): return (self._special[0] < self.ndim) + (self._special[1] < self.ndim)

    nbatch = batch_size
    nchannel = channel_size
    n_space = nspace
    n_special = nspecial

    @property
    def has_batch(self): return self._special[0 if self._batch_first else 1] < self.ndim

    @property
    def has_channel(self): return self._special[1 if self._batch_first else 0] < self.ndim

    @property
    def has_special(self): return self._special != [self.ndim, self.ndim]

    def remove_special_(self):
        self._special = [self.ndim, self.ndim]
        self._batch_first = True
        return self

    def copy(self): return Size(self)

    def __add__(self, other: [tuple, 'Size']):
        if not isinstance(other, tuple):
            raise TypeError("Only Size+tuple is available for Size as a python tuple, "
                            "please use size << 1 to increase the size numerically. ")
        if not isinstance(other, Size):
            return Size(tuple(self) + other).special_from_(self)

        ndim = self.ndim + other.ndim
        batch_dim = channel_dim = ndim
        if self.has_batch: batch_dim = self.batch_dimension
        if other.has_batch:
            if batch_dim < ndim: raise TypeError("Batch dimension conflict in addition. ")
            batch_dim = self.ndim + other.batch_dimension
        if self.has_channel: channel_dim = self.channel_dimension
        if other.has_channel:
            if channel_dim < ndim: raise TypeError("Channel dimension conflict in addition. ")
            channel_dim = self.ndim + other.channel_dimension

        return Size(tuple(self) + tuple(other), batch_dim=batch_dim, channel_dim=channel_dim)

    def __radd__(self, other: [tuple, 'Size']):
        if not isinstance(other, tuple):
            raise TypeError("Only Size+tuple is available for Size as a python tuple, "
                            "please use size << 1 to increase the size numerically. ")
        if not isinstance(other, Size):
            other = Size(other)
        return other + self
    __iadd__ = __add__

    def __mul__(self, value: INT):
        if not isinstance(value, INT):
            raise TypeError("Only Size*int is available for Size as a python tuple, "
                            "please use size ** 1 to do the multiply numerically. ")
        return Size(tuple(self) * value).special_from_(self)
    __imul__ = __rmul__ = __mul__

    @staticmethod
    def __op__(a: [INT, tuple, 'Size'], b: [INT, tuple, 'Size'], *, op):

        def getvalue(x, y, r):
            if x == -2: return y
            if y == -2: return x
            if x == -1 or y == -1: return -1
            if r: z = op(y, x)
            else: z = op(x, y)
            if z < 0: raise Size.NegSizeError
            return INT(z)

        LengthError = TypeError("Operation only apply for Sizes of the same length, "
                                "please consider to identify the batch/channel dimension or use + to concatenate. ")

        rev = False
        if isinstance(a, NUM):
            a, b = b, a
            rev = True
        if isinstance(a, NUM): raise TypeError("'__op__' for Size do not take two numbers. ")
        if isinstance(b, NUM): return Size(getvalue(x, b, rev) for x in a).special_from_(a)
        if isinstance(a, tuple):
            a, b = b, a
            rev = True
        if isinstance(a, tuple):
            if len(a) == len(b): return Size(getvalue(x, y, rev) for x, y in zip(a, b))
            raise LengthError
        if isinstance(b, tuple):
            if len(a) == len(b): return Size(getvalue(x, y, rev) for x, y in zip(a, b)).special_from_(a)
            if a.space == len(b): b = a.insert_special_to_tuple(b, -2)
        
        # Now deal with two Sizes
        if a.ndim == b.nspace:
            a, b = b, a
            rev = True
        if a.ndim == b.nspace:
            if len(a) == len(b): return Size(getvalue(x, y, rev) for x, y in zip(a, b))
            raise LengthError
        if a.nspace == b.ndim: b = a.insert_special_to_tuple(b, -2)
        if a.ndim == b.ndim:
            if a._special == b._special: return Size(getvalue(x, y, rev) for x, y in zip(a, b)).add_special_from_(a).add_special_from_(b)
            raise TypeError("Only Sizes with same batch/channel dimensions can be operated. ")
        else: raise LengthError

    def __lshift__(self, other: [INT, tuple, 'Size']): return self.__op__(self, other, op=lambda x, y: x + y)
    __ilshift__ = __rlshift__ = __lshift__

    def __rshift__(self, other: [INT, tuple, 'Size']): return Size.__op__(self, other, op=lambda x, y: x - y)

    def __rrshift__(self, other: [INT, tuple, 'Size']): return Size.__op__(other, self, op=lambda x, y: x - y)
    __irshift__ = __rshift__

    def __pow__(self, other: [INT, tuple, 'Size']): return Size.__op__(self, other, op=lambda x, y: x * y)
    __ipow__ = __rpow__ = __pow__

    def __floordiv__(self, other: [INT, tuple, 'Size']): return Size.__op__(self, other, op=lambda x, y: x // y)

    def __rfloordiv__(self, other: [INT, tuple, 'Size']): return Size.__op__(other, self, op=lambda x, y: x // y)
    __ifloordiv__ = __floordiv__

    def __xor__(self, other: [tuple, 'Size']):

        if not isinstance(other, Size): other = Size(other)

        if self.ndim == 0: return (Size(1) * other.ndim).special_from_(other), other
        elif other.ndim == 0: return self, (Size(1) * self.ndim).special_from_(self)

        if self.ndim == other.ndim:
            if self.nspecial == 0: return self.special_from_(other), other
            elif other.nspecial == 0: return self, other.special_from_(self)

        if self.nspace == other.nspace:
            if self.nspecial == 0: self = other.insert_special_to_tuple(self, 1)
            elif other.nspecial == 0: other = self.insert_special_to_tuple(other, 1)

        if self.nspecial == other.nspecial:
            if self._batch_first != other._batch_first:
                raise RuntimeError(f"Sizes {self} and {other} with opposite batch-channel order cannot be expand together.")

            self_len1 = self._special[0]
            other_len1 = other._special[0]
            self_len2 = self._special[1] - self._special[0] - 1
            other_len2 = other._special[1] - other._special[0] - 1
            self_len3 = self.ndim - self._special[1] - 1
            other_len3 = other.ndim - other._special[1] - 1
            len1 = MAX(self_len1, other_len1)
            len2 = MAX(self_len2, other_len2)
            len3 = MAX(self_len3, other_len3)
            tup_self = self.tuple()
            tup_other = other.tuple()
            exp_self = (
                (1,) * (len1 - self_len1) + tup_self[:self_len1+1] + 
                (1,) * (len2 - self_len2) + tup_self[self_len1+1:self_len1+self_len2+2] + 
                (1,) * (len3 - self_len3) + tup_self[self_len1+self_len2+2:]
            )
            exp_other = (
                (1,) * (len1 - other_len1) + tup_other[:other_len1+1] + 
                (1,) * (len2 - other_len2) + tup_other[other_len1+1:other_len1+other_len2+2] + 
                (1,) * (len3 - other_len3) + tup_other[other_len1+other_len2+2:]
            )
            if self._batch_first: batch_dim, channel_dim = len1, len1 + len2 + 1
            else: batch_dim, channel_dim = len1 + len2 + 1, len1
            exp_self = Size(exp_self, batch_dim=batch_dim, channel_dim=channel_dim)
            exp_other = Size(exp_other, batch_dim=batch_dim, channel_dim=channel_dim)
            return exp_self, exp_other

        if self.has_batch and other.has_batch:
            lp_self, lp_other = self[:self.batch_dimension] ^ other[:other.batch_dimension]
            rp_self, rp_other = self[self.batch_dimension+1:] ^ other[other.batch_dimension+1:]
            return (
                lp_self + Size([self.batch_size]) + rp_self,
                lp_other + Size([other.batch_size]) + rp_other
            )

        if self.has_channel and other.has_channel:
            lp_self, lp_other = self[:self.channel_dimension] ^ other[:other.channel_dimension]
            rp_self, rp_other = self[self.channel_dimension+1:] ^ other[other.channel_dimension+1:]
            return (
                lp_self + Size({self.channel_size}) + rp_self,
                lp_other + Size({other.channel_size}) + rp_other
            )

        raise RuntimeError("Unexpected error occurs in sizes expanding, please contact developers for more information. ")
    __ixor__ = __rxor__ = __xor__

    def __getitem__(self, k):
        if isinstance(k, INT): return super().__getitem__(k)
        return Size(*self.python_repr[k])

    @property
    def python_repr(self):
        args = list(self)
        batch_dim = self.batch_dimension
        if batch_dim < self.ndim: args[batch_dim] = [args[batch_dim]]
        channel_dim = self.channel_dimension
        if channel_dim < self.ndim: args[channel_dim] = {args[channel_dim]}
        return tuple(args)

    def __str__(self):
        rep = self.python_repr
        if len(rep) == 1: rep = str(rep).replace(',', '')
        return f"torchplus.Size{rep}"
    __repr__ = __str__

class Tensor(torch.Tensor):

    init = False

    @staticmethod
    def _make_subclass(cls, data, auto_device=_auto_device, requires_grad=False, device=None):
        to_device = None
        if device is None and auto_device: to_device = Device
        if device is not None: to_device = device
        if to_device is not None and to_device != data.device: data = data.to(to_device)
        self = torch.Tensor._make_subclass(cls, data, requires_grad)
        if isinstance(data, Tensor): return self.special_from_(data)
        return self

    def __new__(cls, *args, **kwargs):

        to_device = kwargs.get('device', Device if kwargs.get('auto_device', _auto_device) else None)

        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, INT): pass
            elif isinstance(arg, torch.Tensor) and (to_device is None or arg.device == to_device):
                self = arg.as_subclass(Tensor)
                self.requires_grad = kwargs.get('requires_grad', arg.requires_grad)
                return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))
            elif hasattr(arg, 'shape') or isinstance(arg, list):
                if not isinstance(arg, torch.Tensor):
                    if to_device is None: arg = torch.as_tensor(arg)
                    else: arg = torch.as_tensor(arg, device=to_device)
                rg = kwargs.get('requires_grad', arg.requires_grad)
                if to_device is None or arg.device == to_device:
                    self = super()._make_subclass(cls, arg, rg)
                else:
                    self = super()._make_subclass(cls, arg.to(to_device), rg)
                self.special_from_(arg)
                return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))
            elif isinstance(arg, tuple): args = arg
            else: raise TypeError("Unrecognized initialization of 'Tensor'. ")

        # args is a tuple of integers (or a Size)
        if to_device is None: self = super().__new__(cls, *args, **kwargs)
        else: self = super().__new__(cls, *args, device=to_device, **kwargs)
        self.special_from_(args)
        return self.set_special_(kwargs.get('batch_dim', None), kwargs.get('channel_dim', None))

    def requires_grad_(self, mode=True):
        self.requires_grad = mode
        return self

    def refine_names(self, *args, kwargs):
        self.has_names = True
        return super().refine_names(*args, **kwargs)

    @property
    def ishape(self): return super().shape

    @property
    def shape(self):
        shape = Size(super().shape)
        shape.special_from_(self)
        if self.has_names():
            if not shape.has_batch:
                isbatch = [('batch' in x.lower()) if x else x for x in self.names]
                if ANY(isbatch):
                    ibatch = isbatch.index(True)
            if not shape.has_channel:
                ischannel = [('channel' in x.lower()) if x else x for x in self.names]
                if ANY(ischannel):
                    ichannel = ischannel.index(True)
        return shape

    def rename(self, *args, **kwargs):
        ibatch = ichannel = None
        for i, n in enumerate(args):
            if not isinstance(n, str): continue
            elif 'batch' in x.lower():
                if ibatch is not None:
                    raise TypeError("Multiple batch dimensions not supported. ")
                ibatch = i
            elif 'channel' in x.lower():
                if ichannel is not None:
                    raise TypeError("Multiple channel dimensions not supported. ")
                ichannel = i
            if ibatch is not None and ichannel is not None: break
        return super().rename(*args, **kwargs)

    def refine_names(self, *args):
        ibatch = ichannel = None
        for i, n in enumerate(args):
            if not isinstance(n, str): continue
            elif 'batch' in x.lower():
                if ibatch is not None:
                    raise TypeError("Multiple batch dimensions not supported. ")
                ibatch = i
            elif 'channel' in x.lower():
                if ichannel is not None:
                    raise TypeError("Multiple channel dimensions not supported. ")
                ichannel = i
            if ibatch is not None and ichannel is not None: break
        return super().refine_names(*args)
        
    @property
    def batch_dimension(self): return self._special[0 if self._batch_first else 1]

    @batch_dimension.setter
    def batch_dimension(self, batch_dim):
        self.set_special_(batch_dim, None)

    def batch_dimension_(self, value):
        self.batch_dimension = value
        return self

    @property
    def batch_size(self): return self.shape.batch_size

    @property
    def channel_dimension(self): return self._special[1 if self._batch_first else 0]

    @channel_dimension.setter
    def channel_dimension(self, channel_dim):
        self.set_special_(None, channel_dim)

    def channel_dimension_(self, value):
        self.channel_dimension = value
        return self

    @property
    def channel_size(self): return self.shape.channel_size

    @property
    def space(self):
        s = self._special
        t = tuple(self.ishape)
        return t[:s[0]] + t[s[0]+1:s[1]] + t[s[1]+1:]

    @property
    def nele(self): return super().numel()
    def numel(self): return super().numel()

    @property
    def n_dim(self): return self.ndim

    @property
    def nspace(self): return self.ndim if not self.init else (self.ndim - (self._special[0] < self.ndim) - (self._special[1] < self.ndim))

    @property
    def nspecial(self): return 0 if not self.init else ((self._special[0] < self.ndim) + (self._special[1] < self.ndim))

    nbatch = batch_size
    nchannel = channel_size
    n_ele = nele
    n_space = nspace
    n_special = nspecial

    @property
    def has_batch(self): return self.init and self._special[0 if self._batch_first else 1] < self.ndim

    @property
    def has_channel(self): return self.init and self._special[1 if self._batch_first else 0] < self.ndim

    @property
    def has_special(self): return self.init and self._special != [self.ndim, self.ndim]

    @property
    def special(self): return None if not self.init else [x for x in self._special if x < self.ndim]

    def special_from_(self, other=None):

        if isinstance(other, Size) or isinstance(other, Tensor) and other.init:
            self._special = [self.ndim if x == other.ndim else x for x in other._special]
            self._batch_first = other._batch_first
        else:
            self._special = [self.ndim, self.ndim]
            self._batch_first = True

        self.init = True
        return self

    def add_special_from_(self, other=None):

        if isinstance(other, Size):

            doit = False
            batch_dim = None
            channel_dim = None

            if self.batch_dimension == self.ndim:
                batch_dim = other.batch_dimension
                if batch_dim != self.ndim:
                    doit = True
            if self.channel_dimension == self.ndim:
                channel_dim = other.channel_dimension
                if channel_dim != self.ndim:
                    doit = True
            if doit: self.set_special_(batch_dim, channel_dim)

        self.init = True
        return self

    def set_special_(self, batch_dim=None, channel_dim=None, *, special=[], bf=True):

        if batch_dim is None and channel_dim is None:
            if not special:
                if not self.init: return self.special_from_()
                else: return self
            a, b = special
            if not isinstance(a, INT): a = self.ndim
            if not isinstance(b, INT): b = self.ndim
            if a < 0: a += self.ndim
            if b < 0: b += self.ndim
            if not 0 <= a <= self.ndim or not 0 <= b <= self.ndim:
                raise TypeError(f"Special dimension should be a dimension index which is smaller than {self.ndim}. ")
            if not self.init: self._batch_first = bf
            if a < b:
                self._special = [a, b]
            else:
                self._special = [b, a]
                self._batch_first = not self._batch_first
            self.init = True
            return self

        if batch_dim is not None:
            if not isinstance(batch_dim, INT): batch_dim = self.ndim
            if batch_dim < 0: batch_dim = batch_dim + self.ndim
            if not 0 <= batch_dim <= self.ndim:
                raise TypeError(f"batch_dimension should be a dimension index which is smaller than {self.ndim}. ")
        elif self.init:
            batch_dim = MIN(self.batch_dimension, self.ndim)
        else:
            batch_dim = self.ndim

        if channel_dim is not None:
            if not isinstance(channel_dim, INT): channel_dim = self.ndim
            if channel_dim < 0: channel_dim = channel_dim + self.ndim
            if not 0 <= channel_dim <= self.ndim:
                raise TypeError(f"channel_dimension should be a dimension index which is smaller than {self.ndim}. ")
        elif self.init:
            channel_dim = MIN(self.channel_dimension, self.ndim)
        else:
            channel_dim = self.ndim

        if batch_dim < channel_dim:
            self._batch_first = True
            self._special = [batch_dim, channel_dim]
        elif channel_dim < batch_dim:
            self._batch_first = False
            self._special = [channel_dim, batch_dim]
        elif batch_dim < self.ndim:
            raise ValueError(f"special dimensions can not be the same: {batch_dim} and {channel_dim}. ")
        else:
            self._batch_first = True
            self._special = [channel_dim, channel_dim]

        self.init = True
        return self

    def remove_special_(self):
        self._special = [self.ndim, self.ndim]
        self._batch_first = True
        return self

    def remove_batch_(self):
        return self.set_special_(batch_dim=self.ndim)

    def remove_channel_(self):
        return self.set_special_(channel=self.ndim)

    def normalize_(self):
        m, M = self.min(), self.max()
        if m == M:
            if M >= 1: return self.one_()
            if m <= 0: return self.zero_()
            return self
        self.sub_(m)
        self.div_(M-m)
        return self

    def normalize(self):
        m, M = self.min(), self.max()
        if m == M:
            if M >= 1: return ones_like(self)
            if m <= 0: return zeros_like(self)
            return self
        return (self - m) / (M - m)

    for func in tf_list:
        exec(f"def {func}(*args, **kwargs): return tp.{func}(*args, **kwargs)")

    def tensor(self): return super()._make_subclass(torch.Tensor, self, self.requires_grad)
    def numpy(self): return super(torch.Tensor, self.cpu().detach()).numpy()

    def one_(self): return self.zero_().add_(1)
    
    def autodevice_(self):
        if _auto_device and self.device != Device: return self.to(Device)
        return self

    def device_(self, device):
        if device != self.device: return self.to(device)
        return self

    def size(self, *k: [INT, str]):
        if len(k) == 0:
            return self.shape
        i = [(self.names.index(x) if x in self.names else None) if isoftype(x, str) else x for x in k]
        if None in i:
            return super().size(k[i.index(None)])
        if len(i) == 1:
            return self.ishape[i[0]]
        return tuple(self.ishape[x] for x in i)

    def squeeze(self, *dims: int, dim=None):
        if len(dims) > 0 and dim is not None:
            raise TypeError("squeeze function only accept either argument or positional argument. But both are given")
        if len(dims) == 1 and isinstance(dims[0], tuple):
            dims = dims[0]
        if dim is None:
            dim = dims
        if not isinstance(dim, tuple):
            dim = (dim,)
        if len(dim) == 0: dim = tuple(i for i, x in enumerate(self.ishape) if x == 1)[::-1]
        a, b = self._special
        ndim = self.ndim
        bf = self._batch_first
        for d in dim:
            if isinstance(d, list): d = d[0]
            if isinstance(d, set): d = d.pop()
            self = super(Tensor, self).squeeze(d)
            if a == d: a = ndim
            if 0 <= d <= a or d + ndim <= a: a -= 1
            if b == d: b = ndim
            if 0 <= d <= b or d + ndim <= b: b -= 1
            ndim -= 1
        return self.as_subclass(Tensor).set_special_(special=[a, b], bf=bf)

    def squeeze_(self, *dims:int, dim=None):
        if len(dims) > 0 and dim is not None:
            raise TypeError("squeeze_ function only accept either argument or positional argument. But both are given")
        if len(dims) == 1 and isinstance(dims[0], tuple):
            dims = dims[0]
        if dim is None:
            dim = dims
        if not isinstance(dim, tuple):
            dim = (dim,)
        if len(dim) == 0: dim = tuple(i for i, x in enumerate(self.ishape) if x == 1)
        a, b = self._special
        ndim = self.ndim
        bf = self._batch_first
        for d in dim:
            if isinstance(d, list): d = d[0]
            if isinstance(d, set): d = d.pop()
            self = super(Tensor, self).squeeze_(d)
            if a == d: a = ndim
            if 0 <= d <= a or d + ndim <= a: a -= 1
            if b == d: b = ndim
            if 0 <= d <= b or d + ndim <= b: b -= 1
            ndim -= 1
        return self.as_subclass(Tensor).set_special_(special=[a, b], bf=bf)

    def unsqueeze(self, *dims: int, dim=None):
        if len(dims) > 0 and dim is not None:
            raise TypeError("unsqueeze function only accept either argument or positional argument. But both are given")
        if len(dims) == 1 and isinstance(dims[0], tuple):
            dims = dims[0]
        if dim is None:
            dim = dims
        if not isinstance(dim, tuple):
            dim = (dim,)
        if len(dim) == 0: dim = (0,)
        a, b = self._special
        ndim = self.ndim
        bf = self._batch_first
        ibatch = ichannel = None
        for d in dim:
            if isinstance(d, list): d = d[0]; ibatch = d
            if isinstance(d, set): d = d.pop(); ichannel = d
            self = super(Tensor, self).unsqueeze(d)
            if 0 <= d <= a or d + ndim <= a: a += 1
            if 0 <= d <= b or d + ndim <= b: b += 1
            ndim += 1

        return self.as_subclass(Tensor).set_special_(special=[a, b], bf=bf).set_special_(batch_dim=ibatch, channel_dim=ichannel)

    def unsqueeze_(self, *dims:int, dim=None):
        if len(dims) > 0 and dim is not None:
            raise TypeError("unsqueeze_ function only accept either argument or positional argument. But both are given")
        if len(dims) == 1 and isinstance(dims[0], tuple):
            dims = dims[0]
        if dim is None:
            dim = dims
        if not isinstance(dim, tuple):
            dim = (dim,)
        if len(dim) == 0: dim = (0,)
        a, b = self._special
        ndim = self.ndim
        bf = self._batch_first
        ibatch = ichannel = None
        for d in dim:
            if isinstance(d, list): d = d[0]; ibatch = d
            if isinstance(d, set): d = d.pop(); ichannel = d
            self = super(Tensor, self).unsqueeze_(d)
            if 0 <= d <= a or d + ndim <= a: a += 1
            if 0 <= d <= b or d + ndim <= b: b += 1
            ndim += 1
        return self.as_subclass(Tensor).set_special_(special=[a, b], bf=bf).set_special_(batch_dim=ibatch, channel_dim=ichannel)

    def expand_to(self, *target):
        with torch._C.DisableTorchFunction():

            if len(target) == 1 and not isinstance(target[0], INT): target = target[0]
            if isinstance(target, torch.Tensor): target = target.shape
            if not isinstance(target, Size): target = Size(target)
            if self.init and self.nspecial == target.nspecial and target._batch_first != self._batch_first:
                if self.nspace == 0 and self.ndim == 2: self = self[::-1]
                else: raise TypeError(f"Batch and channel order not matched for {self.shape} and {target}")
                
            axis_map = list(RANGE(self.ndim))
            p = 0
            for i, s in enumerate(self.shape):
                if i == self.batch_dimension:
                    axis_map[i] = target.batch_dimension
                    p = target.batch_dimension + 1
                elif i == self.channel_dimension:
                    axis_map[i] = target.channel_dimension
                    p = target.channel_dimension + 1
                elif s in (1, -1):
                    axis_map[i] = p
                    p += 1
                else:
                    while p < target.ndim and target[p] not in (s, -1): p += 1
                    axis_map[i] = p
                    p += 1
                if p >= target.ndim  + 1: raise TypeError(f"Unable to expand sizes {self.shape} to {target}. ")
            res = self.unsqueeze_to(target, axis=axis_map).repeat(tuple(1 if i in axis_map else (x if x >= 0 else 1) for i, x in enumerate(target)))

        return res.as_subclass(Tensor).special_from_(target)

    def unsqueeze_to(self, *target, axis: list):
        with torch._C.DisableTorchFunction():

            if len(target) == 1 and not isinstance(target[0], INT): target = target[0]
            if isinstance(target, torch.Tensor): target = target.shape
            if not isinstance(target, Size): target = Size(target)

            new_size = list(target)
            for i in RANGE(len(new_size)):
                if i not in axis or self.ishape[axis.index(i)] in (1, -1):
                    new_size[i] = 1
            res = self.view(*new_size)

        return res.as_subclass(Tensor).special_from_(target)

    def multiple(self, num, dim=0):
        if isinstance(dim, list): d = dim[0]
        elif isinstance(dim, set): d = dim.pop(); dim = {d}
        else: d = dim
        return self.unsqueeze(dim).repeat((1,) * d + (num,) + (1,) * (self.ndim - d))

    def sample(self, dim: INT = None, number: INT = 1, random: bool = True) -> 'Tensor':
        """
        sample(self, dim: int = self.batch_dimension, numbder: int = 1, random: bool = True) -> Tensor

        Sample a few subspaces from a given dimension.
        data.sample(2, 1, random=False) is equivalant to data[:, :, 0, ...].
        """
        if dim is None or isinstance(dim, list) and dim == []: dim = self.batch_dimension
        if dim is None or isinstance(dim, dict) and dim == {}: dim = self.channel_dimension
        if dim < 0: dim += self.ndim
        if dim is None: raise TypeError("Argument 'dim' needed for sampling Tensors with no special dimensions. ")
        if number < 1: raise TypeError("Argument 'number' for sampling Tensors can not be smaller than 1. ")
        sample_indices = [slice(None)] * self.dim()
        if self.shape[dim] < number: raise TypeError(f"Too many elements needed to be sampled from dimension {dim}")
        if random:
            import random
            samples = random.sample(RANGE(self.shape[dim]), k = number)
        else: samples = list(RANGE(number))
        if len(samples) != 1:
            sample_indices[dim] = samples
            output_tensor = self[tuple(sample_indices)].as_subclass(Tensor).special_from_(self)
            output_tensor.indices = samples
            return output_tensor
        else:
            sample_indices[dim] = samples[0]
            output_tensor = self[tuple(sample_indices)].as_subclass(Tensor).set_special_(special=[x - 1 if x > dim else (self.ndim-1 if x == dim else x) for x in self._special], bf=self._batch_first)
            output_tensor.indices = samples
            return output_tensor

    def pick(self, index: INT, dim: INT):
        """
        pick(self, dim, index) -> Tensor

        pick one of the item on dimension `dim` for big tensors. 
        data.pick(4, 2) is equivalent to data[:, :, 4]
        """
        if dim < 0: dim += self.ndim
        return self[(slice(None),) * dim + (index,)]

    def split(self, sec=1, dim=None, squeeze=False):
        """
        pick(self, dim, index) -> Tensor

        pick one of the item on dimension `dim` for big tensors. 
        data.pick(4, 2) is equivalent to data[:, :, 4]
        """
        if dim is None: dim = self.channel_dimension
        if sec == 1 or isinstance(sec, (tuple, list)) and all(x == 1 for x in sec):
            if squeeze: return tuple(x.as_subclass(Tensor).special_from_(self).squeeze(dim) for x in super().split(sec, dim))
        return tuple(x.as_subclass(Tensor).special_from_(self) for x in super().split(sec, dim))

    def mvdim(self, dim1: INT, dim2: INT):
        """
        mvdim(self, dim1, dim2) -> Tensor

        move dim1 to dim2(specified in the targeting size)
        data of size (2, 3, 4, 5) can be transform to (2, 4, 5, 3) by data.mvdim(1, -1) or data.mvdim(1, 3)
        """
        if dim1 < 0: dim1 += self.ndim
        if dim2 < 0: dim2 += self.ndim

        with torch._C.DisableTorchFunction():
            if dim1 == dim2: res = self
            elif dim1 < dim2: res = self.unsqueeze(dim2+1).transpose(dim1, dim2+1).squeeze(dim1)
            else: res = self.unsqueeze(dim2).transpose(dim1+1, dim2).squeeze(dim1+1)

        return res.as_subclass(Tensor).set_special_(special=[dim2 if x == dim1 else (
            x if x > dim2 and x > dim1 or x < dim1 and x < dim2 
            else (x + 1 if dim1 > dim2 else x - 1)) for x in self._special])

    def mvdim_(self, dim1: INT, dim2: INT):
        """
        In-place operation for mvdim
        """
        if dim1 < 0: dim1 += self.ndim
        if dim2 < 0: dim2 += self.ndim

        with torch._C.DisableTorchFunction():
            if dim1 == dim2: res = self
            elif dim1 < dim2: res = self.unsqueeze_(dim2+1).transpose_(dim1, dim2+1).squeeze_(dim1)
            else: res = self.unsqueeze_(dim2).transpose_(dim1+1, dim2).squeeze_(dim1+1)

        return res.as_subclass(Tensor).special_from_(special=[dim2 if x == dim1 else (
            x if x > dim2 and x > dim1 or x < dim1 and x < dim2 
            else (x + 1 if dim1 > dim2 else x - 1)) for x in self._special])

    def tobytes(self): return self.numpy().tobytes()

    def cat(self, other, dim=0):
        return torch.cat((self, other), dim=dim)

    def stack(self, other, dim=0):
        return torch.stack((self, other), dim=dim)

    @property
    def T(self: 'Tensor') -> 'Tensor':
        if not self.has_special: return super().T
        s = self._special

        with torch._C.DisableTorchFunction():
            permute_dim = tuple(RANGE(s[0]))[::-1] + (s[0],) + tuple(RANGE(s[0]+1, s[1]))[::-1] + (s[1],) * (s[1] != self.ndim) + tuple(RANGE(s[1]+1, self.ndim))[::-1]
        
        return self.permute(*permute_dim)

    def t(self) -> 'Tensor':
        return self.T

    def t_(self) -> 'Tensor':
        """
        t_() -> Tensor

        In-place version of :meth:`~Tensor.t`
        """
        if not self.has_special: return super().t_()
        s = self._special
        with torch._C.DisableTorchFunction():
            for i in RANGE(s[0] // 2):
                self.transpose_(i, s[0] - i - 1)
            for i in RANGE((s[1] - s[0] - 1) // 2):
                self.transpose_(s[0] + i + 1, s[1] - i - 1)
            for i in RANGE((self.ndim - s[1] - 1) // 2):
                self.transpose_(s[1] + i + 1, self.ndim - i - 1)
        return self

    def __getattr__(self, key):
        if not self.init:
            if key == '_special': return [self.ndim, self.ndim]
            elif key == '_batch_first': return True
        try:
            return super().__getattr__(key)
        except:
            raise AttributeError("{} is not a method/attribute of Tensor, the most similar name is {}".format(key, vector(dir(self)).fuzzy_search(key, 3)))

    def __matmul__(self, other, **kwargs):
        if isinstance(other, torch.Tensor) and self.has_special or isinstance(other, Tensor) and other.has_special:
            a, b = self.shape[:-2] ^ other.shape[:-2]
            with torch._C.DisableTorchFunction():
                res = super(torch.Tensor, self.view(a + tuple(self.shape[-2:]))).__matmul__(other.view(b + tuple(other.shape[-2:])))
            return res.as_subclass(Tensor).special_from_(self.shape[:-2])
        return super().__matmul__(other, **kwargs).as_subclass(Tensor).special_from_()

    def __op__(self, opname, other, **kwargs):
        if self.has_special or isinstance(other, Tensor) and other.has_special:
            if not isinstance(other, Tensor): other = tensor(other)
            a, b = self.shape ^ other.shape
            with torch._C.DisableTorchFunction():
                res = getattr(super(Tensor, self.view(a).as_subclass(Tensor)), opname)(other.view(b))
            return res.as_subclass(Tensor).special_from_(a)
        return getattr(super(), opname)(other, **kwargs).as_subclass(Tensor).special_from_()

    for op in '''
    add iadd radd
    sub isub rsub
    mul imul rmul
    div idiv rdiv
    pow ipow rpow
    mod imod rmod
    truediv itruediv rtruediv
    floordiv ifloordiv rfloordiv
    eq ieq req
    ne ine rne
    or ior ror
    and iand rand
    xor ixor rxor
    lt le gt ge
    '''.split():
        exec(f"def __{op}__(self, *args, **kwargs): return self.__op__('__{op}__', *args, **kwargs)")

    ###### old operation code ######
    #    if isinstance(other, torch.Tensor):
    #        other = Tensor(other)
    #        if self.dim() == other.dim():
    #            return super().__add__(other)
    #        elif self.dim() < other.dim():
    #            return self.expand_as(other).__add__(other)
    #        else:
    #            return super().__add__(other.expand_as(self))
    #    return super().__add__(other)
    #################################

    def __repr__(self, *args, **kwargs):
        string = super().__repr__(*args, **kwargs)
        if 'shape=' not in string:
            string = string.rstrip(')') + f', shape={self.shape})'
        return string.replace("tensor", "Tensor")

    def __str__(self, *args, **kwargs):
        string = super().__str__(*args, **kwargs)
        if 'shape=' not in string:
            string = string.rstrip(')') + f', shape={self.shape})'
        return string.replace("tensor", "Tensor")

    def __hash__(self): return super().__hash__()

    @staticmethod
    def __torch_function_convert__(ret, cls):
        with torch._C.DisableTorchFunction():
            if isinstance(ret, torch.Tensor):
                ret = ret.as_subclass(cls)
            if isinstance(ret, (tuple, list)):
                # Also handles things like namedtuples
                ret = type(ret)(Tensor.__torch_function_convert__(r, cls) for r in ret)
            return ret

    @staticmethod
    def __torch_function_collect__(r, c):
        if isinstance(r, (tuple, list)):
            for x in r: Tensor.__torch_function_collect__(x, c)
        if isinstance(r, Tensor) and not r.init: c.append(r)

    @staticmethod
    def __torch_function_convert_apply__(ret, apply, cls):

        with torch._C.DisableTorchFunction():

            if isinstance(ret, Tensor) and not ret.init:
                if cls != Tensor: ret = ret.as_subclass(cls)
                apply(ret)
                return ret

            if isinstance(ret, torch.Tensor):
                ret = ret.as_subclass(cls)
                return ret

            if isinstance(ret, (tuple, list)):
                # Also handles things like namedtuples
                return type(ret)(Tensor.__torch_function_convert_apply__(r, apply, cls) for r in ret)

            if 'tuple_iterator' in str(type(ret)):
                # And things iterator like Generators
                def out():
                    for r in ret:
                        if isinstance(r, torch.Tensor):
                            r = r.as_subclass(cls)
                            apply(r)
                            yield r
                return out()

            return ret

    @classmethod
    def __torch_function_ele_wise_func__(cls, func, types, args=(), kwargs=None):
        """
        FOR: tensor where
        __add__ __iadd__ __radd__
        __sub__ __isub__ __rsub__
        __mul__ __imul__ __rmul__
        __div__ __idiv__ __rdiv__
        __pow__ __ipow__ __rpow__
        __mod__ __imod__ __rmod__
        __truediv__ __itruediv__ __rtruediv__
        __floordiv__ __ifloordiv__ __rfloordiv__
        __eq__ __ieq__ __req__
        __ne__ __ine__ __rne__
        __or__ __ior__ __ror__
        __and__ __iand__ __rand__
        __xor__ __ixor__ __rxor__
        __lt__ __le__ __gt__ __ge__
        """
        self = args[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(self)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_resizing_func__(cls, func, types, args=(), kwargs=None):
        "FOR: reshape view zeros ones rand randn"
        dims = args[1:]
        if len(dims) == 1: dims = dims[0]
        if not isinstance(dims, Size): dims = Size(dims)
        args = args[:1] + tuple(dims)

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(dims)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_resizing_as_func__(cls, func, types, args=(), kwargs=None):
        "FOR: reshape_as view_as unsqueeze_to expand_to"
        dims = args[-1].shape

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(dims)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_randint_func__(cls, func, types, args=(), kwargs=None):
        "FOR: randint"
        dims = Size(args[3])
        args = args[:3] + (dims,)

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(dims)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_transpose_func__(cls, func, types, args=(), kwargs=None):
        "FOR: transpose transpose_"
        self = args[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            dim1 = kwget(kwargs, 'dim0', args[1])
            dim2 = kwget(kwargs, 'dim1', args[2])
            a, b = self._special
            if a == dim1: a = dim2
            elif a == dim2: a = dim1
            if b == dim1: b = dim2
            elif b == dim2: b = dim1
            r.set_special_(special=[a, b], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_permute_func__(cls, func, types, args=(), kwargs=None):
        "FOR: permute"
        self = args[0]
        dims = args[1:]
        if len(dims) == 1 and isinstance(dims[0] (list, tuple)): dims = dims[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        d = lambda i: dims[i] if i < len(dims) else self.ndim
        def apply(r):
            r.set_special_(special=[d(self._special[0]), d(self._special[1])], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_matmul_func__(cls, func, types, args=(), kwargs=None):
        "FOR: mm bmm smm matmul __matmul__"
        self, other = args[:2]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(self.shape[:-1] + other.shape[-1:])

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_addmm_func__(cls, func, types, args=(), kwargs=None):
        "FOR: addmm addbmm saddmm"
        _, self, weight = args[:3]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(self.shape[:-1] + other.shape[-1:])

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_reducing_func__(cls, func, types, args=(), kwargs=None):
        "FOR: cummin cummax cumsum cumprod sum prod min max mean std argmin argmax"
        func_name = str(func).split(' of ')[0].split()[-1].strip("'")
        mkwargs = kwargs if kwargs is not None else {}
        if len(args) == 1 and 'dim' not in mkwargs and func_name in ('sum', 'prod', 'mean', 'std', 'cumsum', 'cumprod'):
            self = args[0]
            s = self._special
            dims = list(RANGE(s[0])) + list(RANGE(s[0]+1, s[1])) + list(RANGE(s[1]+1, self.ndim))
            args += (dims,)

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            if len(args) > 1:
                self = args[0]
                dims = mkwargs.get('dim', args[1:])
                if len(dims) == 1 and isinstance(dims[0], (list, tuple)): dims = dims[0]
                if len(dims) == 0: r.special_from_()
                else: r.set_special_(special=[x - len([d for d in dims if 0 <= d < x or d + self.ndim < x]) if x not in dims else r.ndim for x in self._special], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_expanding_func__(cls, func, types, args=(), kwargs=None):
        "FOR: unsqueeze unsqueeze_ squeeze squeeze_"
        # self = args[0]
        # dims = kwget(kwargs, 'dim', args[1:])
        # if len(dims) == 1 and isinstance(dims[0], tuple): dims = dims[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r): ...
            # a, b = self._special
            # ndim = self.ndim
            # for d in dims:
            #     if 0 <= d <= a or d + ndim <= a: a += 1
            #     if 0 <= d <= b or d + ndim <= b: b += 1
            #     ndim += 1
            # r.set_special_(special=[a, b], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_flatten_func__(cls, func, types, args=(), kwargs=None):
        "FOR: flatten flatten_"
        dims = kwget(kwargs, 'dim', args[1:])
        if isinstance(dims, tuple) and len(dims) <= 0:
            dims = kwget(kwargs, 'start_dim', 0), kwget(kwargs, 'end_dim', -1)

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            self = args[0]
            if len(dims) == 1:
                dim = dims[0]
                if dim < 0: dim += self.ndim
                r.set_special_(special=[r.ndim if x >= dim else x for x in self._special], bf=self._batch_first)
            else:
                dim1, dim2 = dims
                if dim1 < 0: dim1 += self.ndim
                if dim2 < 0: dim2 += self.ndim
                r.set_special_(special=[r.ndim if dim1 <= x <= dim2 or x > r.ndim else x for x in self._special], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_getitem_func__(cls, func, types, args=(), kwargs=None):
        "FOR: __getitem__ __iter__"
        if len(args) == 1: self, dims = args[0], 0
        else: self, dims = args

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

            ndim = self.ndim

            if not isinstance(dims, tuple): dims = (dims,)
            types = [type(x) for x in dims]
            if type(...) in types:
                if types.count(type(...)) > 1:
                    raise TypeError("")
                lp = dims[:types.index(type(...))]
                rp = dims[types.index(type(...))+1:]
            else:
                lp = dims
                rp = tuple()
            offset = ndim - len(rp)
            isnormal = ''.join('1' if issubclass(t, (INT, slice, type(...))) else '0' for t in types)
            if '1' in isnormal.strip('1') or '0' not in isnormal: index = -1
            else: index = isnormal.index('0')

        def apply(r):
            a, b = self._special
            rdim = r.ndim
            offset2 = rdim - ndim + len(lp) + len(rp) - len([d for d in lp + rp if isinstance(d, slice)])
            if a < len(lp) and not isinstance(lp[a], slice): a = None
            elif offset <= a < ndim and not isinstance(rp[a - offset], slice): a = None
            if b < len(lp) and not isinstance(lp[b], slice): b = None
            elif offset <= b < ndim and not isinstance(rp[b - offset], slice): b = None
            if a is not None and a < ndim:
                a += offset2 if a > index else 0
                a -= len([d for d in RANGE(len(lp)) if (0 <= d < a or d + ndim < a) and not isinstance(lp[d], slice)])
                a -= len([d for d in RANGE(offset, ndim) if (0 <= d < a or d + ndim < a) and not isinstance(rp[d-offset], slice)])
            elif a is not None: a = rdim
            if b is not None and b < ndim:
                b += offset2 if b > index else 0
                b -= len([d for d in RANGE(len(lp)) if (0 <= d < b or d + ndim < b) and not isinstance(lp[d], slice)])
                b -= len([d for d in RANGE(offset, ndim) if (0 <= d < b or d + ndim < b) and not isinstance(rp[d-offset], slice)])
            elif b is not None: b = rdim
            r.set_special_(special=[a, b], bf=self._batch_first)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_concatenate_func__(cls, func, types, args=(), kwargs=None):
        "FOR: cat concatenate"
        self = args[0]
        if isinstance(self, tuple): self = self[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(self)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_stack_func__(cls, func, types, args=(), kwargs=None):
        "FOR: stack"
        self = args[0]
        if isinstance(self, tuple): self = self[0]
        dim = kwget(kwargs, 'dim', args[-1] if isinstance(args[-1], INT) else 0)

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(special=[x + 1 if x >= dim else x for x in self._special])

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    @classmethod
    def __torch_function_default_func__(cls, func, types, args=(), kwargs=None):
        "FOR: all the remainings"
        self = args[0]

        with torch._C.DisableTorchFunction():
            ret = super().__torch_function__(func, types, args, kwargs)
            if isinstance(ret, type(NotImplemented)):
                raise NotImplementedError(f"{func} for {args} is not implemented. ")

        def apply(r):
            r.special_from_(self)

        return Tensor.__torch_function_convert_apply__(ret, apply, cls)

    def __torch_function_keys__(func):
        return func.__func__.__doc__.split('\n')[0].split(':')[-1].strip().split()

    __torch_function_map__ = {k: '__torch_function_ele_wise_func__' for k in __torch_function_keys__(__torch_function_ele_wise_func__)}
    __torch_function_map__.update({k: '__torch_function_resizing_func__' for k in __torch_function_keys__(__torch_function_resizing_func__)})
    __torch_function_map__.update({k: '__torch_function_resizing_as_func__' for k in __torch_function_keys__(__torch_function_resizing_as_func__)})
    __torch_function_map__.update({k: '__torch_function_randint_func__' for k in __torch_function_keys__(__torch_function_randint_func__)})
    __torch_function_map__.update({k: '__torch_function_transpose_func__' for k in __torch_function_keys__(__torch_function_transpose_func__)})
    __torch_function_map__.update({k: '__torch_function_permute_func__' for k in __torch_function_keys__(__torch_function_permute_func__)})
    __torch_function_map__.update({k: '__torch_function_matmul_func__' for k in __torch_function_keys__(__torch_function_matmul_func__)})
    __torch_function_map__.update({k: '__torch_function_addmm_func__' for k in __torch_function_keys__(__torch_function_addmm_func__)})
    __torch_function_map__.update({k: '__torch_function_reducing_func__' for k in __torch_function_keys__(__torch_function_reducing_func__)})
    __torch_function_map__.update({k: '__torch_function_expanding_func__' for k in __torch_function_keys__(__torch_function_expanding_func__)})
    __torch_function_map__.update({k: '__torch_function_flatten_func__' for k in __torch_function_keys__(__torch_function_flatten_func__)})
    __torch_function_map__.update({k: '__torch_function_getitem_func__' for k in __torch_function_keys__(__torch_function_getitem_func__)})
    __torch_function_map__.update({k: '__torch_function_concatenate_func__' for k in __torch_function_keys__(__torch_function_concatenate_func__)})
    __torch_function_map__.update({k: '__torch_function_stack_func__' for k in __torch_function_keys__(__torch_function_stack_func__)})

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):

        if len(args) == 0: return super().__torch_function__(func, types, args, kwargs)

        sfunc = str(func)
        if sfunc.startswith('<attribute') or sfunc.startswith('<property'):
            return super().__torch_function__(func, types, args, kwargs)

        func_name = sfunc.split(' of ')[0].split(' at ')[0].split()[-1].strip("'").split('.')[-1]
        if func_name in ('__get__', '__set__', '__delete__'):
            return super().__torch_function__(func, types, args, kwargs)

        self = args[0]
        types = tuple(cls if t in [torch.nn.Parameter, tp.nn.Parameter] else t for t in types)
        torch_func_name = Tensor.__torch_function_map__.get(func_name, None)
        if isinstance(self, Tensor) and self.init and self.has_special: pass
        elif torch_func_name in ('__torch_function_resizing_func__', '__torch_function_resizing_as_func__', '__torch_function_randint_func__'): pass
        else:
            with torch._C.DisableTorchFunction():
                ret = super().__torch_function__(func, types, args, kwargs)
                collection = []
            def apply(r): r.special_from_()
            return Tensor.__torch_function_convert_apply__(ret, apply, cls)

        if torch_func_name is None: return Tensor.__torch_function_default_func__(func, types, args, kwargs)
        else: return eval(f"Tensor.{torch_func_name}")(func, types, args, kwargs)


for func in '''
zeros ones
rand randn
'''.split():
    __all__.extend([func, func+'_like'])
    exec(f"""

def {func}(*args, **kwargs):

    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, torch.Tensor):
            return torch.{func}_like(arg, **kwargs).as_subclass(Tensor).special_from_(arg).autodevice_()
        elif isinstance(arg, (tuple, list)):
            size = Size(*arg)
            return torch.{func}(size, **kwargs).as_subclass(Tensor).special_from_(size).autodevice_()
        else:
            return torch.{func}(*args, **kwargs).as_subclass(Tensor).special_from_().autodevice_()

    size = Size(*args)
    return torch.{func}(size, **kwargs).as_subclass(Tensor).special_from_(size).autodevice_()

def {func}_like(tensor, **kwargs):
    return {func}(tensor, **kwargs)
""")

for func in '''
range arange
'''.split():
    __all__.append(func)
    exec(f"""
def {func}(*args, **kwargs):
    return torch.{func}(*args, **kwargs).as_subclass(Tensor).special_from_().autodevice_()
""")

class _Randint:

    def __init__(self):
        self.range = (0, 2)

    def __getitem__(self, t):
        if len(t) == 0: t = (0, 2)
        elif len(t) == 1: t = (0, t[0])
        elif len(t) > 2: raise TypeError(f"Please use randint[lower, upper] to specify the range with upper end excluded. ")
        self.range = t
        return self

    def __call__(self, *size, **kwargs):
        return torch.randint(self.range[0], self.range[1], Size(size), **kwargs).as_subclass(Tensor).special_from_(size).autodevice_()

class _Randint_like:

    def __init__(self):
        self.range = (0, 2)

    def __getitem__(self, t):
        if len(t) == 0: t = (0, 2)
        elif len(t) == 1: t = (0, t[0])
        elif len(t) > 2: raise TypeError(f"Please use randint_like[lower, upper] to specify the range with upper end excluded. ")
        self.range = t
        return self

    def __call__(self, data, **kwargs):
        return torch.randint_like(data, self.range[0], self.range[1], **kwargs).as_subclass(Tensor).special_from_(data.shape).autodevice_()

randint = _Randint()
randint_like = _Randint_like()
__all__.extend(["randint", "randint_like"])

__all__.extend(["eye", "cat", "stack", "t", "unsqueeze", "tensor"])

def eye(*size: tuple, **kwargs):
    if len(size) == 1 and isinstance(size, (tuple, list)): size = size[0]

    size = Size(size)
    if size.nspace < 1: raise TypeError("Empty size not valid for 'eye'. ")
    if size.nspace == 1: size = size + (size.space[0],)
    if size.nspace > 2: raise TypeError("No more than 2-D is allowed for 'eye'. ")
    n = MIN(*size.space)
    s = [slice(None)] * size.ndim
    for i in RANGE(size.ndim):
        if i not in size.special:
            s[i] = torch.arange(n)
    out = zeros(size, **kwargs)
    out[tuple(s)] = 1
    return out

def cat(*list_of_tensors, dim=None, **kwargs):
    if dim is None:
        if len(list_of_tensors) > 1 and not isinstance(list_of_tensors[-1], Tensor):
            dim = list_of_tensors[-1]
            list_of_tensors =list_of_tensors[:-1]
        else: dim = 0
    if len(list_of_tensors) == 1 and isinstance(list_of_tensors[0], (tuple, list)):
        list_of_tensors = list_of_tensors[0]
    ibatch = ichannel = None
    if isinstance(dim, list): dim = dim[0]; ibatch = dim
    if isinstance(dim, set): dim = dim.pop(); ichannel = dim
    self = [t for t in list_of_tensors if isinstance(t, Tensor)]
    if len(self) == 0: return torch.cat(list_of_tensors, dim, **kwargs).as_subclass(Tensor).special_from_().set_special_(batch_dim=ibatch, channel_dim=ichannel)
    return torch.cat(list_of_tensors, dim, **kwargs).as_subclass(Tensor).special_from_(self[0]).set_special_(batch_dim=ibatch, channel_dim=ichannel)

def stack(*list_of_tensors, dim=None, **kwargs):
    if dim is None:
        if len(list_of_tensors) > 1 and not isinstance(list_of_tensors[-1], Tensor):
            dim = list_of_tensors[-1]
            list_of_tensors =list_of_tensors[:-1]
        else: dim = 0
    if len(list_of_tensors) == 1 and isinstance(list_of_tensors[0], (tuple, list)):
        list_of_tensors = list_of_tensors[0]
    ibatch = ichannel = None
    if isinstance(dim, list): dim = dim[0]; ibatch = dim
    if isinstance(dim, set): dim = dim.pop(); ichannel = dim
    self = [t for t in list_of_tensors if isinstance(t, Tensor)]
    if len(self) == 0: return torch.stack(list_of_tensors, dim, **kwargs).as_subclass(Tensor).special_from_().set_special_(batch_dim=ibatch, channel_dim=ichannel)
    return torch.stack(list_of_tensors, dim, **kwargs).as_subclass(Tensor).set_special_(special=[x + 1 if x >= dim else x for x in self[0]._special], bf=self[0]._batch_first).set_special_(batch_dim=ibatch, channel_dim=ichannel)

def t(tensor: Array.Torch):
    if isinstance(tensor, Tensor): return tensor.T
    else: return torch.t(tensor).as_subclass(Tensor).special_from_(tensor)

def squeeze(tensor, *args, **kwargs):
    if isinstance(tensor, Tensor): return tensor.squeeze(*args, **kwargs)
    else: return torch.squeeze(tensor, *args, **kwargs).as_subclass(Tensor).special_from_(tensor)

def unsqueeze(tensor, *args, **kwargs):
    if isinstance(tensor, Tensor): return tensor.unsqueeze(*args, **kwargs)
    else: return torch.unsqueeze(tensor, *args, **kwargs).as_subclass(Tensor).special_from_(tensor)

def tensor(data, *, dtype=None, device=None, requires_grad=False, pin_memory=False):
    if device is None and _auto_device is True:
        device = Device
    if isinstance(data, Tensor):
        return data.clone()
    if isinstance(data, torch.Tensor):
        return data.as_subclass(Tensor).special_from_()
    return torch.tensor(data, dtype=dtype, device=device, requires_grad=requires_grad, pin_memory=pin_memory).as_subclass(Tensor).special_from_()

# print(torch_type_list - (torch_type_list - globals()))
# print(torch_dtype_list - (torch_dtype_list - globals()))
# print(torch_func_list - (torch_func_list - globals()))

for key in torch_dtype_list:
    if not (key in __all__ or key in globals()):
        exec(f"{key} = torch.{key}")
        __all__.append(key)

for key in torch_type_list:
    if not (key in __all__ or key in globals()):
        exec(f"{key} = torch.{key}")
        __all__.append(key)

for key in dir(torch):
    if key.startswith("_"):
        continue
    if inspect.isclass(eval(f"torch.{key}")):
        continue
    if (key in __all__ or key in globals()) and key not in {"typename"}:
        continue
    if key in torch_func_list:
        # functions
        exec(f"{key} = torch.{key}")
        __all__.append(key)




from pycamia import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = 'micomputing',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.1.31',
    contact = 'bertiezhou@163.com',
    keywords = ['medical image', 'image registration', 'image similarities'],
    description = "'micomputing' is a package for medical image computing. ",
    requires = ['numpy', 'torch>=1.5.1', 'pycamia', 'pyoverload', 'batorch']
).check()

from .stdio import NII, DCM #*
from .sim import MutualInformation, NormalizedMutualInformation, KLDivergence, CorrelationOfLocalEstimation, NormalizedVectorInformation, Cos2Theta, SumSquaredDifference, MeanSquaredErrors, PeakSignalToNoiseRatio, CrossEntropy, CrossCorrelation, NormalizedCrossCorrelation, StructuralSimilarity, Dice, LabelDice #*











































































































